"""Base class for embedders and models."""

import abc
from pathlib import Path
from typing import Any, TypeVar, Union

import geopandas as gpd
import pandas as pd

from srai.constants import GEOMETRY_COLUMN

try:  # pragma: no cover
    from pytorch_lightning import LightningModule

except ImportError:
    from srai.embedders._pytorch_stubs import LightningModule


class Model(LightningModule):  # type: ignore
    """Class for model based on LightningModule."""

    def get_config(self) -> dict[str, Any]:
        """Get model config."""
        model_config = {
            k: v
            for k, v in vars(self).items()
            if k[0] != "_"
            and k
            not in (
                "training",
                "prepare_data_per_node",
                "allow_zero_length_dataloader_with_multiple_devices",
            )
        }

        return model_config

    def save(self, path: Union[Path, str]) -> None:
        """
        Save the model to a directory.

        Args:
            path (Path): Path to the directory.
        """
        import torch

        torch.save(self.state_dict(), path)

    @classmethod
    def load(cls, path: Union[Path, str], **kwargs: Any) -> "Model":
        """
        Load model from a file.

        Args:
            path (Union[Path, str]): Path to the file.
            **kwargs (dict): Additional kwargs to pass to the model constructor.
        """
        import torch

        if isinstance(path, str):
            path = Path(path)

        model = cls(**kwargs)
        model.load_state_dict(torch.load(path))
        return model


class Embedder(abc.ABC):
    """Abstract class for embedders."""

    @abc.abstractmethod
    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:  # pragma: no cover
        """
        Embed regions using features.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding and geometry index for each region in regions_gdf.

        Raises:
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        raise NotImplementedError

    # @abc.abstractmethod
    # def save(self, path: Union[Path, str]) -> None:
    #     """
    #     Save the embedder to a directory.

    #     Args:
    #         path (Path): Path to the directory.
    #     """
    #     raise NotImplementedError

    # @classmethod
    # @abc.abstractmethod
    # def load(cls, path: Union[Path, str]) -> "Embedder":
    #     """
    #     Load the embedder from a directory.

    #     Args:
    #         path (Path): Path to the directory.

    #     Returns:
    #         Embedder: The loaded embedder.
    #     """
    #     raise NotImplementedError

    def _validate_indexes(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> None:
        if regions_gdf.index.name is None:
            raise ValueError("regions_gdf must have a named index.")

        if features_gdf.index.name is None:
            raise ValueError("features_gdf must have a named index.")

        if not isinstance(joint_gdf.index, pd.MultiIndex):
            raise ValueError(
                f"joint_gdf.index must be of type pandas.MultiIndex, not {type(joint_gdf.index)}"
            )

        if len(joint_gdf.index.names) != 2:
            raise ValueError(
                f"joint_gdf.index must have 2 levels, has {len(joint_gdf.index.names)}"
            )

        if regions_gdf.index.name != joint_gdf.index.names[0]:
            raise ValueError(
                f"Name of regions_gdf.index ({regions_gdf.index.name}) must be equal to the name of"
                f" the 1st level of joint_gdf.index ({joint_gdf.index.names[0]})"
            )

        if features_gdf.index.name != joint_gdf.index.names[1]:
            raise ValueError(
                f"Name of features_gdf.index ({features_gdf.index.name}) must be equal to the name"
                f" of the 2nd level of joint_gdf.index ({joint_gdf.index.names[1]})"
            )

    def _remove_geometry_if_present(self, data: gpd.GeoDataFrame) -> pd.DataFrame:
        if GEOMETRY_COLUMN in data.columns:
            data = data.drop(columns=GEOMETRY_COLUMN)
        return pd.DataFrame(data)


ModelT = TypeVar("ModelT", bound=Model)



"""
Contextual Count Embedder.

This module contains contextual count embedder implementation from ARIC@SIGSPATIAL 2021 paper [1].

References:
    1. https://doi.org/10.1145/3486626.3493434
    1. https://arxiv.org/abs/2111.00990
"""

from collections.abc import Iterator
from typing import Optional, Union

import geopandas as gpd
import numpy as np
import numpy.typing as npt
import pandas as pd

from srai.embedders.count_embedder import CountEmbedder
from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter, OsmTagsFilter
from srai.neighbourhoods import Neighbourhood
from srai.neighbourhoods._base import IndexType


class ContextualCountEmbedder(CountEmbedder):
    """ContextualCountEmbedder."""

    def __init__(
        self,
        neighbourhood: Neighbourhood[IndexType],
        neighbourhood_distance: int,
        concatenate_vectors: bool = False,
        expected_output_features: Optional[
            Union[list[str], OsmTagsFilter, GroupedOsmTagsFilter]
        ] = None,
        count_subcategories: bool = False,
    ) -> None:
        """
        Init ContextualCountEmbedder.

        Args:
            neighbourhood (Neighbourhood[T]): Neighbourhood object used to get neighbours for
                the contextualization.
            neighbourhood_distance (int): How many neighbours levels should be included in
                the embedding.
            concatenate_vectors (bool, optional): Whether to sum all neighbours into a single vector
                with the same width as `CountEmbedder`, or to concatenate them to the wide format
                and keep all neighbour levels separate. Defaults to False.
            expected_output_features
                (Union[List[str], OsmTagsFilter, GroupedOsmTagsFilter], optional):
                The features that are expected to be found in the resulting embedding.
                If not None, the missing features are added and filled with 0.
                The unexpected features are removed. The resulting columns are sorted accordingly.
                Defaults to None.
            count_subcategories (bool, optional): Whether to count all subcategories individually
                or count features only on the highest level based on features column name.
                Defaults to False.

        Raises:
            ValueError: If `neighbourhood_distance` is negative.
        """
        super().__init__(expected_output_features, count_subcategories)

        self.neighbourhood = neighbourhood
        self.neighbourhood_distance = neighbourhood_distance
        self.concatenate_vectors = concatenate_vectors

        if self.neighbourhood_distance < 0:
            raise ValueError("Neighbourhood distance must be positive.")

    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Embed a given GeoDataFrame.

        Creates region embeddings by counting the frequencies of each feature value and applying
        a contextualization based on neighbours of regions. For each region, features will be
        altered based on the neighbours either by adding averaged values dimished based on distance,
        or by adding new separate columns with neighbour distance postfix.
        Expects features_gdf to be in wide format with each column being a separate type of
        feature (e.g. amenity, leisure) and rows to hold values of these features for each object.
        The rows will hold numbers of this type of feature in each region. Numbers can be
        fractional because neighbourhoods are averaged to represent a single value from
        all neighbours on a given level.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding for each region in regions_gdf.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        counts_df = super().transform(regions_gdf, features_gdf, joint_gdf)

        result_df: pd.DataFrame
        if self.concatenate_vectors:
            result_df = self._get_concatenated_embeddings(counts_df)
        else:
            result_df = self._get_squashed_embeddings(counts_df)

        return result_df

    def _get_squashed_embeddings(self, counts_df: pd.DataFrame) -> pd.DataFrame:
        """
        Generate embeddings for regions by summing all neighbourhood levels.

        Creates embedding by getting an average of a neighbourhood at a given distance and adding it
        to the base values with weight equal to 1 / (distance + 1) squared. This way, farther
        neighbourhoods have lower impact on feature values.

        Args:
            counts_df (pd.DataFrame): Calculated features from CountEmbedder.

        Returns:
            pd.DataFrame: Embedding for each region in regions_gdf with number of features equal to
                the same as returned by the CountEmbedder.
        """
        base_columns = list(counts_df.columns)

        result_array = counts_df.values.astype(float)

        for distance, averaged_values in self._get_averaged_values_for_distances(counts_df):
            result_array += averaged_values / ((distance + 1) ** 2)

        return pd.DataFrame(data=result_array, index=counts_df.index, columns=base_columns)

    def _get_concatenated_embeddings(self, counts_df: pd.DataFrame) -> pd.DataFrame:
        """
        Generate embeddings for regions by concatenating different neighbourhood levels.

        Creates embedding by getting an average of a neighbourhood at a given distance and adding
        those features as separate columns with a postfix added to feature name. This way,
        all neighbourhoods can be analyzed separately, but number of columns grows linearly with
        a distance.

        Args:
            counts_df (pd.DataFrame): Calculated features from CountEmbedder.

        Returns:
            pd.DataFrame: Embedding for each region in regions_gdf with number of features equal to
                a number of features returned by the CountEmbedder multiplied
                by (neighbourhood distance + 1).
        """
        base_columns = list(counts_df.columns)
        number_of_base_columns = len(base_columns)
        columns = [
            f"{column}_{distance}"
            for distance in range(self.neighbourhood_distance + 1)
            for column in base_columns
        ]

        result_array = np.zeros(shape=(len(counts_df.index), len(columns)))
        result_array[:, 0:number_of_base_columns] = counts_df.values

        for distance, averaged_values in self._get_averaged_values_for_distances(counts_df):
            result_array[
                :, number_of_base_columns * distance : number_of_base_columns * (distance + 1)
            ] = averaged_values

        return pd.DataFrame(data=result_array, index=counts_df.index, columns=columns)

    def _get_averaged_values_for_distances(
        self, counts_df: pd.DataFrame
    ) -> Iterator[tuple[int, npt.NDArray[np.float32]]]:
        """
        Generate averaged values for neighbours at given distances.

        Function will yield tuples of distances and averaged values arrays
        calculated based on neighbours at a given distance.

        Distance 0 is skipped.
        If embedder has `neighbourhood_distance` set to 0, nothing will be returned.

        Args:
            counts_df (pd.DataFrame): Calculated features from CountEmbedder.

        Yields:
            Iterator[Tuple[int, npt.NDArray[np.float32]]]: Iterator of distances and values.
        """
        if self.neighbourhood_distance == 0:
            return

        number_of_base_columns = len(counts_df.columns)

        for distance in range(1, self.neighbourhood_distance + 1):
            neighbours_series = counts_df.index.map(
                lambda region_id, neighbour_distance=distance: counts_df.index.intersection(
                    self.neighbourhood.get_neighbours_at_distance(
                        region_id, neighbour_distance, include_center=False
                    )
                ).values
            )
            if len(neighbours_series) == 0:
                continue

            averaged_values_stacked = np.stack(
                neighbours_series.map(
                    lambda region_ids: (
                        np.nan_to_num(np.nanmean(counts_df.loc[region_ids].values, axis=0))
                        if len(region_ids) > 0
                        else np.zeros((number_of_base_columns,))
                    )
                ).values
            )

            yield distance, averaged_values_stacked



"""
This module contains neighbourhood calculation methods.

Some embedding methods require a neighbourhood to be defined. This module contains
neighbourhood calculation methods - both dedicated to a specific regionalization method
and general ones.
"""
from ._base import Neighbourhood
from .adjacency_neighbourhood import AdjacencyNeighbourhood
from .h3_neighbourhood import H3Neighbourhood

__all__ = ["Neighbourhood", "AdjacencyNeighbourhood", "H3Neighbourhood"]



"""
GTFS Loader.

This module contains GTFS loader which performs time aggregations from timetable data.
It utilizes the `gtfs_kit` library [1]. It was originally created for the purpose of
the gtfs2vec project [2].

References:
    1. https://gitlab.com/mrcagney/gtfs_kit
    2. https://doi.org/10.1145/3486640.3491392
"""

from pathlib import Path
from typing import TYPE_CHECKING

import geopandas as gpd
import pandas as pd
from shapely.geometry import Point

from srai._optional import import_optional_dependencies
from srai.constants import GEOMETRY_COLUMN, WGS84_CRS
from srai.loaders import Loader

if TYPE_CHECKING:  # pragma: no cover
    from gtfs_kit import Feed

GTFS2VEC_DIRECTIONS_PREFIX = "directions_at_"
GTFS2VEC_TRIPS_PREFIX = "trips_at_"


class GTFSLoader(Loader):
    """
    GTFSLoader.

    This loader is capable of reading GTFS feed and calculates time aggregations in 1H slots.
    """

    def __init__(self) -> None:
        """Initialize GTFS loader."""
        import_optional_dependencies(dependency_group="gtfs", modules=["gtfs_kit"])

        self.time_resolution = "1H"

    def load(
        self,
        gtfs_file: Path,
        fail_on_validation_errors: bool = True,
        skip_validation: bool = False,
    ) -> gpd.GeoDataFrame:
        """
        Load GTFS feed and calculate time aggregations for stops.

        Args:
            gtfs_file (Path): Path to the GTFS feed.
            fail_on_validation_errors (bool): Fail if GTFS feed is invalid. Ignored when
                skip_validation is True.
            skip_validation (bool): Skip GTFS feed validation.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with trip counts and list of directions for stops.
        """
        import gtfs_kit as gk

        feed = gk.read_feed(gtfs_file, dist_units="km")

        if not skip_validation:
            self._validate_feed(feed, fail=fail_on_validation_errors)

        trips_df = self._load_trips(feed)
        directions_df = self._load_directions(feed)

        stops_df = feed.stops[["stop_id", "stop_lat", "stop_lon"]].set_index("stop_id")
        stops_df[GEOMETRY_COLUMN] = stops_df.apply(
            lambda row: Point([row["stop_lon"], row["stop_lat"]]), axis=1
        )

        result_gdf = gpd.GeoDataFrame(
            trips_df.merge(stops_df[GEOMETRY_COLUMN], how="inner", on="stop_id"),
            geometry=GEOMETRY_COLUMN,
            crs=WGS84_CRS,
        )

        result_gdf = result_gdf.merge(directions_df, how="left", on="stop_id")

        result_gdf.index.name = None

        return result_gdf

    def _load_trips(self, feed: "Feed") -> pd.DataFrame:
        """
        Load trips from GTFS feed.

        Calculate sum of trips from stop in each time slot.

        Args:
            feed (gk.Feed): GTFS feed.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with trips.
        """
        # FIXME: this takes first wednesday from the feed, may not be the best,
        # but that is what I did in gtfs2vec
        date = feed.get_first_week()[2]
        ts = feed.compute_stop_time_series([date], freq=self.time_resolution)

        records = []

        for idx, row in ts.iterrows():
            h = idx.hour
            for s, n in row["num_trips"].items():
                records.append((s, h, n))

        df = pd.DataFrame(records, columns=["stop_id", "hour", "num_trips"])
        df = df.pivot_table(index="stop_id", columns="hour", values="num_trips", fill_value=0)
        df = df.add_prefix(GTFS2VEC_TRIPS_PREFIX)

        return df

    def _load_directions(self, feed: "Feed") -> gpd.GeoDataFrame:
        """
        Load directions from GTFS feed.

        Create a list of unique directions for each stop and time slot.

        Args:
            feed (gk.Feed): GTFS feed.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with directions.
        """
        df = feed.stop_times.merge(feed.trips, on="trip_id")
        df = df.merge(feed.stops, on="stop_id")

        df = df[df["departure_time"].notna()]

        df["hour"] = df["departure_time"].apply(self._parse_departure_time)

        pivoted = df.pivot_table(
            values="trip_headsign", index="stop_id", columns="hour", aggfunc=set
        )
        pivoted = pivoted.add_prefix(GTFS2VEC_DIRECTIONS_PREFIX)

        return pivoted

    def _validate_feed(self, feed: "Feed", fail: bool = True) -> None:
        """
        Validate GTFS feed.

        Args:
            feed (gk.Feed): GTFS feed.
            fail (bool): Fail if feed is invalid.
        """
        validation_result = feed.validate()

        if (validation_result["type"] == "error").sum() > 0:
            import warnings

            warnings.warn(f"Invalid GTFS feed: \n{validation_result}", RuntimeWarning, stacklevel=2)
            if fail:
                raise ValueError("Invalid GTFS feed.")

    def _parse_departure_time(self, departure_time: str) -> int:
        """
        Parse departure time and extract hour from it.

        In GTFS feed, departure time is in format HH:MM:SS. HH can be greater than 24, so
        we need to parse it to 0-23 range.

        Args:
            departure_time (str): Departure time in format HH:MM:SS.

        Returns:
            int: Departure time in hours.
        """
        return int(departure_time[:2].replace(":", "")) % 24



"""
OpenStreetMap extracts.

This module contains iterators for publically available OpenStreetMap `*.osm.pbf` files
repositories.
"""
import json
import re
from dataclasses import asdict, dataclass
from functools import partial
from math import ceil
from multiprocessing import cpu_count
from pathlib import Path
from typing import Any, Iterable, List, Optional, Set, Union

import geopandas as gpd
import requests
from shapely.geometry import MultiPolygon
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map

from srai._optional import import_optional_dependencies
from srai.constants import WGS84_CRS
from srai.geometry import flatten_geometry

OPENSTREETMAP_FR_POLYGONS_INDEX_URL = "https://download.openstreetmap.fr/polygons"
OPENSTREETMAP_FR_EXTRACTS_INDEX_URL = "https://download.openstreetmap.fr/extracts"
OPENSTREETMAP_FR_INDEX_GDF: Optional[gpd.GeoDataFrame] = None

GEOFABRIK_INDEX_URL = "https://download.geofabrik.de/index-v1.json"
GEOFABRIK_INDEX_GDF: Optional[gpd.GeoDataFrame] = None


@dataclass
class OpenStreetMapExtract:
    """OSM Extract metadata object."""

    id: str
    url: str
    geometry: BaseGeometry


def find_smallest_containing_geofabrik_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
) -> List[OpenStreetMapExtract]:
    """
    Find smallest extracts from Geofabrik that contains given geometry.

    Iterates a geofabrik index and finds smallest extracts that covers a given geometry.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.

    Returns:
        List[GeofabrikExtract]: List of extracts name, URL to download it and boundary polygon.
    """
    global GEOFABRIK_INDEX_GDF  # noqa: PLW0603

    if GEOFABRIK_INDEX_GDF is None:
        GEOFABRIK_INDEX_GDF = _load_geofabrik_index()

    return _find_smallest_containing_extracts(geometry, GEOFABRIK_INDEX_GDF)


def find_smallest_containing_openstreetmap_fr_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
) -> List[OpenStreetMapExtract]:
    """
    Find smallest extracts from OpenStreetMap.fr that contains given polygon.

    Iterates an osm.fr index and finds smallest extracts that covers a given geometry.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.

    Returns:
        List[GeofabrikExtract]: List of extracts name, URL to download it and boundary polygon.
    """
    global OPENSTREETMAP_FR_INDEX_GDF  # noqa: PLW0603

    import_optional_dependencies(dependency_group="osm", modules=["bs4"])

    if OPENSTREETMAP_FR_INDEX_GDF is None:
        OPENSTREETMAP_FR_INDEX_GDF = _load_openstreetmap_fr_index()

    return _find_smallest_containing_extracts(geometry, OPENSTREETMAP_FR_INDEX_GDF)


def _find_smallest_containing_extracts(
    geometry: Union[BaseGeometry, BaseMultipartGeometry],
    polygons_index_gdf: gpd.GeoDataFrame,
    num_of_multiprocessing_workers: int = -1,
    multiprocessing_activation_threshold: Optional[int] = None,
) -> List[OpenStreetMapExtract]:
    """
    Find smallest set of extracts covering a given geometry.

    Iterates a provided extracts index and searches for a smallest set that cover a given geometry.
    It's not guaranteed that this set will be the smallest and there will be no overlaps.

    Geometry will be flattened into singluar geometries if it's `BaseMultipartGeometry`.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        polygons_index_gdf (gpd.GeoDataFrame): Index of available extracts.
        num_of_multiprocessing_workers (int, optional): Number of workers used for multiprocessing.
            Defaults to -1 which results in a total number of available cpu threads.
            `0` and `1` values disable multiprocessing.
            Similar to `n_jobs` parameter from `scikit-learn` library.
        multiprocessing_activation_threshold (int, optional): Number of gometries required to start
            processing on multiple processes. Activating multiprocessing for a small
            amount of points might not be feasible. Defaults to 100.

    Returns:
        List[OpenStreetMapExtract]: List of extracts covering a given geometry.
    """
    if num_of_multiprocessing_workers == 0:
        num_of_multiprocessing_workers = 1
    elif num_of_multiprocessing_workers < 0:
        num_of_multiprocessing_workers = cpu_count()

    if not multiprocessing_activation_threshold:
        multiprocessing_activation_threshold = 100

    unique_extracts_ids: Set[str] = set()

    geometries = flatten_geometry(geometry)

    total_polygons = len(geometries)

    if (
        num_of_multiprocessing_workers > 1
        and total_polygons >= multiprocessing_activation_threshold
    ):
        find_extracts_func = partial(
            _find_smallest_containing_extracts_for_single_geometry,
            polygons_index_gdf=polygons_index_gdf,
        )

        for extract_ids_list in process_map(
            find_extracts_func,
            geometries,
            desc="Finding matching extracts",
            max_workers=num_of_multiprocessing_workers,
            chunksize=ceil(total_polygons / (4 * num_of_multiprocessing_workers)),
        ):
            unique_extracts_ids.update(extract_ids_list)
    else:
        for sub_geometry in tqdm(geometries, desc="Finding matching extracts"):
            unique_extracts_ids.update(
                _find_smallest_containing_extracts_for_single_geometry(
                    sub_geometry, polygons_index_gdf
                )
            )

    extracts_filtered = _filter_extracts(
        geometry,
        unique_extracts_ids,
        polygons_index_gdf,
        num_of_multiprocessing_workers,
        multiprocessing_activation_threshold,
    )

    return extracts_filtered


def _find_smallest_containing_extracts_for_single_geometry(
    geometry: BaseGeometry, polygons_index_gdf: gpd.GeoDataFrame
) -> Set[str]:
    """
    Find smallest set of extracts covering a given singular geometry.

    Args:
        geometry (BaseGeometry): Geometry to be covered.
        polygons_index_gdf (gpd.GeoDataFrame): Index of available extracts.

    Raises:
        RuntimeError: If provided extracts index is empty.
        RuntimeError: If there is no extracts covering a given geometry (singularly or in group).

    Returns:
        Set[str]: Selected extract index string values.
    """
    if polygons_index_gdf is None:
        raise RuntimeError("Extracts index is empty.")

    extracts_ids: Set[str] = set()
    geometry_to_cover = geometry.buffer(0)

    exactly_matching_geometry = polygons_index_gdf[
        polygons_index_gdf.geometry.geom_almost_equals(geometry)
    ]
    if len(exactly_matching_geometry) == 1:
        extracts_ids.add(exactly_matching_geometry.iloc[0].id)
        return extracts_ids

    iterations = 100
    while not geometry_to_cover.is_empty and iterations > 0:
        matching_rows = polygons_index_gdf[
            (~polygons_index_gdf["id"].isin(extracts_ids))
            & (polygons_index_gdf.intersects(geometry_to_cover))
        ]
        if 0 in (len(matching_rows), iterations):
            raise RuntimeError("Couldn't find extracts matching given geometry.")

        smallest_extract = matching_rows.iloc[0]
        geometry_to_cover = geometry_to_cover.difference(smallest_extract.geometry)
        extracts_ids.add(smallest_extract.id)
        iterations -= 1
    return extracts_ids


def _filter_extracts(
    geometry: BaseGeometry,
    extracts_ids: Iterable[str],
    polygons_index_gdf: gpd.GeoDataFrame,
    num_of_multiprocessing_workers: int,
    multiprocessing_activation_threshold: int,
) -> List[OpenStreetMapExtract]:
    """
    Filter a set of extracts to include least overlaps in it.

    Args:
        geometry (Union[BaseGeometry, BaseMultipartGeometry]): Geometry to be covered.
        extracts_ids (Iterable[str]): Group of selected extracts indexes.
        polygons_index_gdf (gpd.GeoDataFrame): Index of available extracts.
        num_of_multiprocessing_workers (int): Number of workers used for multiprocessing.
            Similar to `n_jobs` parameter from `scikit-learn` library.
        multiprocessing_activation_threshold (int): Number of gometries required to start
            processing on multiple processes.

    Raises:
        RuntimeError: If provided extracts index is empty.

    Returns:
        List[OpenStreetMapExtract]: Filtered list of extracts.
    """
    if polygons_index_gdf is None:
        raise RuntimeError("Extracts index is empty.")

    sorted_extracts_gdf = polygons_index_gdf[
        polygons_index_gdf["id"].isin(extracts_ids)
    ].sort_values(by="area", ignore_index=True, ascending=False)

    filtered_extracts: List[OpenStreetMapExtract] = []
    filtered_extracts_ids: Set[str] = set()

    geometries = flatten_geometry(geometry)

    total_geometries = len(geometries)

    if (
        num_of_multiprocessing_workers > 1
        and total_geometries >= multiprocessing_activation_threshold
    ):
        filter_extracts_func = partial(
            _filter_extracts_for_single_geometry,
            sorted_extracts_gdf=sorted_extracts_gdf,
        )

        for extract_ids_list in process_map(
            filter_extracts_func,
            geometries,
            desc="Filtering extracts",
            max_workers=num_of_multiprocessing_workers,
            chunksize=ceil(total_geometries / (4 * num_of_multiprocessing_workers)),
        ):
            filtered_extracts_ids.update(extract_ids_list)
    else:
        for sub_geometry in tqdm(geometries, desc="Filtering extracts"):
            filtered_extracts_ids.update(
                _filter_extracts_for_single_geometry(sub_geometry, sorted_extracts_gdf)
            )

    simplified_extracts_ids = _simplify_selected_extracts(
        filtered_extracts_ids, sorted_extracts_gdf
    )

    for _, extract_row in sorted_extracts_gdf[
        sorted_extracts_gdf["id"].isin(simplified_extracts_ids)
    ].iterrows():
        extract = OpenStreetMapExtract(
            id=extract_row.id,
            url=extract_row["url"],
            geometry=extract_row.geometry,
        )
        filtered_extracts.append(extract)

    return filtered_extracts


def _filter_extracts_for_single_geometry(
    geometry: BaseGeometry, sorted_extracts_gdf: gpd.GeoDataFrame
) -> Set[str]:
    """
    Filter a set of extracts to include least overlaps in it for a single geometry.

    Works by selecting biggest extracts (by area) and not including smaller ones if they don't
    increase a coverage.

    Args:
        geometry (BaseGeometry): Geometry to be covered.
        sorted_extracts_gdf (gpd.GeoDataFrame): Sorted index of available extracts.

    Returns:
        Set[str]: Selected extract index string values.
    """
    polygon_to_cover = geometry.buffer(0)
    filtered_extracts_ids: Set[str] = set()

    polygon_to_cover = geometry.buffer(0)
    for _, extract_row in sorted_extracts_gdf.iterrows():
        if polygon_to_cover.is_empty:
            break

        if extract_row.geometry.disjoint(polygon_to_cover):
            continue

        polygon_to_cover = polygon_to_cover.difference(extract_row.geometry)
        filtered_extracts_ids.add(extract_row.id)

    return filtered_extracts_ids


def _simplify_selected_extracts(
    filtered_extracts_ids: Set[str], sorted_extracts_gdf: gpd.GeoDataFrame
) -> Set[str]:
    simplified_extracts_ids: Set[str] = filtered_extracts_ids.copy()

    matching_extracts = sorted_extracts_gdf[sorted_extracts_gdf["id"].isin(simplified_extracts_ids)]

    simplify_again = True
    while simplify_again:
        simplify_again = False
        extract_to_remove = None
        for extract_id in simplified_extracts_ids:
            extract_geometry = (
                matching_extracts[sorted_extracts_gdf["id"] == extract_id].iloc[0].geometry
            )
            other_geometries = matching_extracts[
                sorted_extracts_gdf["id"] != extract_id
            ].unary_union
            if extract_geometry.covered_by(other_geometries):
                extract_to_remove = extract_id
                simplify_again = True
                break

        if extract_to_remove is not None:
            simplified_extracts_ids.remove(extract_to_remove)

    return simplified_extracts_ids


def _load_geofabrik_index() -> gpd.GeoDataFrame:
    """
    Load available extracts from GeoFabrik download service.

    Returns:
        gpd.GeoDataFrame: Extracts index with metadata.
    """
    save_path = Path("cache/geofabrik_index.geojson")

    if save_path.exists():
        gdf = gpd.read_file(save_path)
    else:
        result = requests.get(
            GEOFABRIK_INDEX_URL,
            headers={"User-Agent": "SRAI Python package (https://github.com/kraina-ai/srai)"},
        )
        parsed_data = json.loads(result.text)
        gdf = gpd.GeoDataFrame.from_features(parsed_data["features"])
        gdf["area"] = gdf.geometry.area
        gdf.sort_values(by="area", ignore_index=True, inplace=True)
        gdf["url"] = gdf["urls"].apply(lambda d: d["pbf"])
        gdf = gdf[["id", "name", "geometry", "area", "url"]]

        save_path.parent.mkdir(parents=True, exist_ok=True)
        gdf.to_file(save_path, driver="GeoJSON")

    return gdf


def _load_openstreetmap_fr_index() -> gpd.GeoDataFrame:
    """
    Load available extracts from OpenStreetMap.fr download service.

    Returns:
        gpd.GeoDataFrame: Extracts index with metadata.
    """
    save_path = Path("cache/osm_fr_index.geojson")

    if save_path.exists():
        gdf = gpd.read_file(save_path)
    else:
        with tqdm() as pbar:
            extracts = _iterate_openstreetmap_fr_index("osm_fr", "/", True, pbar)
        gdf = gpd.GeoDataFrame(
            data=[asdict(extract) for extract in extracts], geometry="geometry"
        ).set_crs(WGS84_CRS)
        gdf["area"] = gdf.geometry.area
        gdf.sort_values(by="area", ignore_index=True, inplace=True)

        save_path.parent.mkdir(parents=True, exist_ok=True)
        gdf.to_file(save_path, driver="GeoJSON")

    return gdf


def _iterate_openstreetmap_fr_index(
    id_prefix: str, directory_url: str, return_extracts: bool, pbar: tqdm
) -> List[OpenStreetMapExtract]:
    """
    Iterate OpenStreetMap.fr extracts service page.

    Works recursively, by scraping whole available directory.

    Args:
        id_prefix (str): Prefix to be applies to extracts names.
        directory_url (str): Directory URL to load.
        return_extracts (bool): Whether to return collected extracts or not.
        pbar (tqdm): Progress bar.

    Returns:
        List[OpenStreetMapExtract]: List of loaded osm.fr extracts objects.
    """
    from bs4 import BeautifulSoup

    pbar.set_description_str(id_prefix)
    extracts = []
    result = requests.get(
        f"{OPENSTREETMAP_FR_EXTRACTS_INDEX_URL}{directory_url}",
        headers={"User-Agent": "SRAI Python package (https://github.com/kraina-ai/srai)"},
    )
    soup = BeautifulSoup(result.text, "html.parser")
    if return_extracts:
        extracts_urls = soup.find_all(string=re.compile("-latest\\.osm\\.pbf$"))
        for extract in extracts_urls:
            link = extract.find_parent("tr").find("a")
            name = link.text.replace("-latest.osm.pbf", "")
            polygon = _parse_polygon_file(f"{directory_url}{name}.poly")
            if polygon is None:
                continue
            extracts.append(
                OpenStreetMapExtract(
                    id=f"{id_prefix}_{name}",
                    url=f"{OPENSTREETMAP_FR_EXTRACTS_INDEX_URL}{directory_url}{link['href']}",
                    geometry=polygon,
                )
            )
            pbar.update()
    directories = soup.find_all(src="/icons/folder.gif")
    for directory in directories:
        link = directory.find_parent("tr").find("a")
        name = link.text.replace("/", "")
        extracts.extend(
            _iterate_openstreetmap_fr_index(
                id_prefix=f"{id_prefix}_{name}",
                directory_url=f"{directory_url}{link['href']}",
                return_extracts=True,
                pbar=pbar,
            )
        )

    return extracts


def _parse_polygon_file(polygon_url: str) -> Optional[MultiPolygon]:
    """
    Parse poly file from URL to geometry.

    Args:
        polygon_url (str): URL to load a poly file.

    Returns:
        Optional[MultiPolygon]: Parsed polygon.
            Empty if request returns 404 not found.
    """
    result = requests.get(
        f"{OPENSTREETMAP_FR_POLYGONS_INDEX_URL}/{polygon_url}",
        headers={"User-Agent": "SRAI Python package (https://github.com/kraina-ai/srai)"},
    )
    if result.status_code == 404:
        return None
    result.raise_for_status()
    poly = parse_poly(result.text.splitlines())
    return poly


def parse_poly(lines: List[str]) -> MultiPolygon:
    """
    Parse an Osmosis polygon filter file.

    Accept a sequence of lines from a polygon file, return a shapely.geometry.MultiPolygon object.
    Based on: https://wiki.openstreetmap.org/wiki/Osmosis/Polygon_Filter_File_Python_Parsing

    http://wiki.openstreetmap.org/wiki/Osmosis/Polygon_Filter_File_Format
    """
    in_ring = False
    coords: List[Any] = []

    for index, line in enumerate(lines):
        if index == 0:
            # first line is junk.
            continue

        elif index == 1:
            # second line is the first polygon ring.
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True

        elif in_ring and line.strip() == "END":
            # we are at the end of a ring, perhaps with more to come.
            in_ring = False

        elif in_ring:
            # we are in a ring and picking up new coordinates.
            ring.append(list(map(float, line.split())))

        elif not in_ring and line.strip() == "END":
            # we are at the end of the whole polygon.
            break

        elif not in_ring and line.startswith("!"):
            # we are at the start of a polygon part hole.
            coords[-1][1].append([])
            ring = coords[-1][1][-1]
            in_ring = True

        elif not in_ring:
            # we are at the start of a polygon part.
            coords.append([[], []])
            ring = coords[-1][0]
            in_ring = True

    return MultiPolygon(coords)



"""
PBF File Handler.

This module contains a handler capable of parsing a PBF file into a GeoDataFrame.
"""
import warnings
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Sequence, Union, cast

import geopandas as gpd
import osmium
import osmium.osm
import shapely.wkb as wkblib
from osmium.osm.types import T_obj
from shapely.geometry import MultiPolygon, Polygon
from shapely.geometry.base import BaseGeometry
from tqdm import tqdm

from srai.constants import FEATURES_INDEX, WGS84_CRS
from srai.loaders.osm_loaders.filters import OsmTagsFilter

if TYPE_CHECKING:
    import os


class PbfFileHandler(osmium.SimpleHandler):  # type: ignore
    """
    PbfFileHandler.

    PBF(Protocolbuffer Binary Format)[1] file handler is a wrapper around
    a `SimpleHandler`[2] from the `pyosmium`[3] library capable of parsing an `*.osm.pbf` file.

    Handler requires tags filter to only unpack required objects and additionally can use
    a geometry to filter only intersecting objects.

    Handler inherits functions from the `SimpleHandler`, such as `apply_file` and `apply_buffer`
    but it's discouraged to use them on your own, and instead use dedicated `get_features_gdf`
    function.

    References:
        1. https://wiki.openstreetmap.org/wiki/PBF_Format
        2. https://docs.osmcode.org/pyosmium/latest/ref_osmium.html#osmium.SimpleHandler
        3. https://osmcode.org/pyosmium/
    """

    _PBAR_FORMAT = "[{}] Parsing pbf file #{}"

    def __init__(
        self,
        tags: Optional[OsmTagsFilter] = None,
        region_geometry: Optional[BaseGeometry] = None,
    ) -> None:
        """
        Initialize PbfFileHandler.

        Args:
            tags (osm_tags_type, optional): A dictionary
                specifying which tags to download.
                The keys should be OSM tags (e.g. `building`, `amenity`).
                The values should either be `True` for retrieving all objects with the tag,
                string for retrieving a single tag-value pair
                or list of strings for retrieving all values specified in the list.
                `tags={'leisure': 'park}` would return parks from the area.
                `tags={'leisure': 'park, 'amenity': True, 'shop': ['bakery', 'bicycle']}`
                would return parks, all amenity types, bakeries and bicycle shops.
                If `None`, handler will allow all of the tags to be parsed. Defaults to `None`.
            region_geometry (BaseGeometry, optional): Region which can be used to filter only
                intersecting OSM objects. Defaults to None.
        """
        super().__init__()
        self.filter_tags = tags
        if self.filter_tags:
            self.filter_tags_keys = set(self.filter_tags.keys())
        else:
            self.filter_tags_keys = set()
        self.region_geometry = region_geometry
        self.wkbfab = osmium.geom.WKBFactory()
        self.features_cache: Dict[str, Dict[str, Any]] = {}
        self.features_count: Optional[int] = None

    def get_features_gdf(
        self, file_paths: Sequence[Union[str, "os.PathLike[str]"]], region_id: str = "OSM"
    ) -> gpd.GeoDataFrame:
        """
        Get features GeoDataFrame from a list of PBF files.

        Function parses multiple PBF files and returns a single GeoDataFrame with parsed
        OSM objects.

        This function is a dedicated wrapper around the inherited function `apply_file`.

        Args:
            file_paths (Sequence[Union[str, os.PathLike[str]]]): List of paths to `*.osm.pbf`
                files to be parsed.
            region_id (str, optional): Region name to be set in progress bar.
                Defaults to "OSM".

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with OSM features.
        """
        self._clear_cache()
        if self.features_count is None:
            self._count_features(file_paths, region_id)

        with tqdm(desc="Parsing pbf file", total=self.features_count) as self.pbar:
            for path_no, path in enumerate(file_paths):
                self.path_no = path_no + 1
                description = self._PBAR_FORMAT.format(region_id, str(self.path_no))
                self.pbar.set_description(description)
                self.apply_file(path)
            if self.features_cache:
                features_gdf = (
                    gpd.GeoDataFrame(data=self.features_cache.values())
                    .set_crs(WGS84_CRS)
                    .set_index(FEATURES_INDEX)
                )
            else:
                features_gdf = gpd.GeoDataFrame(
                    index=gpd.pd.Index(name=FEATURES_INDEX, data=[]), crs=WGS84_CRS, geometry=[]
                )

        self._clear_cache()
        return features_gdf

    def node(self, node: osmium.osm.Node) -> None:
        """
        Implementation of the required `node` function.

        See [1] for more information.

        Args:
            node (osmium.osm.Node): Node to be parsed.

        References:
            1. https://docs.osmcode.org/pyosmium/latest/ref_osm.html#osmium.osm.Node
        """
        if self.counting_features:
            self._count_feature()
        else:
            self._parse_osm_object(
                osm_object=node, osm_type="node", parse_to_wkb_function=self.wkbfab.create_point
            )

    def way(self, way: osmium.osm.Way) -> None:
        """
        Implementation of the required `way` function.

        See [1] for more information.

        Args:
            way (osmium.osm.Way): Way to be parsed.

        References:
            1. https://docs.osmcode.org/pyosmium/latest/ref_osm.html#osmium.osm.Way
        """
        if self.counting_features:
            self._count_feature()
        else:
            self._parse_osm_object(
                osm_object=way, osm_type="way", parse_to_wkb_function=self.wkbfab.create_linestring
            )

    def area(self, area: osmium.osm.Area) -> None:
        """
        Implementation of the required `area` function.

        See [1] for more information.

        Args:
            area (osmium.osm.Area): Area to be parsed.

        References:
            1. https://docs.osmcode.org/pyosmium/latest/ref_osm.html#osmium.osm.Area
        """
        if self.counting_features:
            self._count_feature()
        else:
            self._parse_osm_object(
                osm_object=area,
                osm_type="way" if area.from_way() else "relation",
                parse_to_wkb_function=self.wkbfab.create_multipolygon,
                osm_id=area.orig_id(),
            )

    def _clear_cache(self) -> None:
        """Clear memory from accumulated features."""
        self.features_cache.clear()
        self.features_count = None

    def _count_features(
        self, file_paths: Sequence[Union[str, "os.PathLike[str]"]], region_id: str = "OSM"
    ) -> None:
        with tqdm(desc=f"[{region_id}] Counting pbf features") as self.pbar:
            self.counting_features = True
            self.features_count = 0
            for path in file_paths:
                self.apply_file(path)
            self.pbar.update(n=self.features_count % 100_000)
            self.counting_features = False

    def _count_feature(self) -> None:
        self.features_count = cast(int, self.features_count) + 1
        if self.features_count % 100_000 == 0:
            self.pbar.update(n=100_000)

    def _parse_osm_object(
        self,
        osm_object: osmium.osm.OSMObject[T_obj],
        osm_type: str,
        parse_to_wkb_function: Callable[..., str],
        osm_id: Optional[int] = None,
    ) -> None:
        """Parse OSM object into a feature with geometry and tags if it matches given criteria."""
        self.pbar.update(n=1)

        if osm_id is None:
            osm_id = osm_object.id

        full_osm_id = f"{osm_type}/{osm_id}"

        matching_tags = self._get_matching_tags(osm_object)
        if matching_tags:
            geometry = self._get_osm_geometry(osm_object, parse_to_wkb_function)
            self._add_feature_to_cache(
                full_osm_id=full_osm_id, matching_tags=matching_tags, geometry=geometry
            )

    def _get_matching_tags(self, osm_object: osmium.osm.OSMObject[T_obj]) -> Dict[str, str]:
        """
        Find matching tags between provided filter and currently parsed OSM object.

        If tags filter is `None`, it will copy all tags from the OSM object.
        """
        matching_tags: Dict[str, str] = {}

        if self.filter_tags:
            for tag_key in self.filter_tags_keys:
                if tag_key in osm_object.tags:
                    object_tag_value = osm_object.tags[tag_key]
                    filter_tag_value = self.filter_tags[tag_key]
                    if (
                        (isinstance(filter_tag_value, bool) and filter_tag_value)
                        or (
                            isinstance(filter_tag_value, str)
                            and object_tag_value == filter_tag_value
                        )
                        or (
                            isinstance(filter_tag_value, list)
                            and object_tag_value in filter_tag_value
                        )
                    ):
                        matching_tags[tag_key] = object_tag_value
        else:
            for tag in osm_object.tags:
                matching_tags[tag.k] = tag.v

        return matching_tags

    def _get_osm_geometry(
        self, osm_object: osmium.osm.OSMObject[T_obj], parse_to_wkb_function: Callable[..., str]
    ) -> BaseGeometry:
        """Get geometry from currently parsed OSM object."""
        geometry = None
        try:
            wkb = parse_to_wkb_function(osm_object)
            geometry = wkblib.loads(wkb, hex=True)
        except Exception as ex:
            message = str(ex)
            warnings.warn(message, RuntimeWarning, stacklevel=2)

        return geometry

    def _add_feature_to_cache(
        self, full_osm_id: str, matching_tags: Dict[str, str], geometry: Optional[BaseGeometry]
    ) -> None:
        """
        Add OSM feature to cache or update existing one based on ID.

        Some of the `way` features are parsed twice, in form of `LineStrings` and `Polygons` /
        `MultiPolygons`. Additional check ensures that a closed geometry will be always preffered
        over a `LineString`.
        """
        if geometry is not None and self._geometry_is_in_region(geometry):
            if full_osm_id not in self.features_cache:
                self.features_cache[full_osm_id] = {
                    FEATURES_INDEX: full_osm_id,
                    "geometry": geometry,
                }

            if isinstance(geometry, (Polygon, MultiPolygon)):
                self.features_cache[full_osm_id]["geometry"] = geometry

            self.features_cache[full_osm_id].update(matching_tags)

    def _geometry_is_in_region(self, geometry: BaseGeometry) -> bool:
        """Check if OSM geometry intersects with provided region."""
        return self.region_geometry is None or geometry.intersects(self.region_geometry)



"""Base class for OSM loaders."""

import abc
from typing import Iterable, Optional, Union, cast

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry.base import BaseGeometry
from tqdm import tqdm

from srai._typing import is_expected_type
from srai.constants import WGS84_CRS
from srai.loaders import Loader
from srai.loaders.osm_loaders.filters import (
    GroupedOsmTagsFilter,
    OsmTagsFilter,
    merge_grouped_osm_tags_filter,
)


def prepare_area_gdf_for_loader(
    area: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]
) -> gpd.GeoDataFrame:
    """
    Prepare an area for the loader.

    Loader expects a GeoDataFrame input, but users shouldn't be limited by this requirement.
    All Shapely geometries will by transformed into GeoDataFrame with proper CRS.

    Args:
        area (Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]):
            Area to be parsed into GeoDataFrame.

    Returns:
        gpd.GeoDataFrame: Sanitized GeoDataFrame.
    """
    if isinstance(area, gpd.GeoDataFrame):
        # Return a GeoDataFrame with changed CRS
        return area.to_crs(WGS84_CRS)
    elif isinstance(area, gpd.GeoSeries):
        # Create a GeoDataFrame with GeoSeries
        return gpd.GeoDataFrame(geometry=area, crs=WGS84_CRS)
    elif isinstance(area, Iterable):
        # Create a GeoSeries with a list of geometries
        return prepare_area_gdf_for_loader(gpd.GeoSeries(area, crs=WGS84_CRS))
    # Wrap a single geometry with a list
    return prepare_area_gdf_for_loader([area])


class OSMLoader(Loader, abc.ABC):
    """Abstract class for loaders."""

    OSM_FILTER_GROUP_COLUMN_NAME = "osm_group_"

    @abc.abstractmethod
    def load(
        self,
        area: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame],
        tags: Union[OsmTagsFilter, GroupedOsmTagsFilter],
    ) -> gpd.GeoDataFrame:  # pragma: no cover
        """
        Load data for a given area.

        Args:
            area (Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]):
                Shapely geometry with the area of interest.
            tags (Union[OsmTagsFilter, GroupedOsmTagsFilter]): OSM tags filter.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with the downloaded data.
        """
        raise NotImplementedError

    def _prepare_area_gdf(
        self, area: Union[BaseGeometry, Iterable[BaseGeometry], gpd.GeoSeries, gpd.GeoDataFrame]
    ) -> gpd.GeoDataFrame:
        return prepare_area_gdf_for_loader(area)

    def _merge_osm_tags_filter(
        self, tags: Union[OsmTagsFilter, GroupedOsmTagsFilter]
    ) -> OsmTagsFilter:
        """
        Merge OSM tags filter into `osm_tags_type` type.

        Optionally merges `GroupedOsmTagsFilter` into `OsmTagsFilter` to allow loaders to load all
        defined groups during single operation.

        Args:
            tags (Union[OsmTagsFilter, GroupedOsmTagsFilter]): OSM tags filter definition.

        Raises:
            AttributeError: When provided tags don't match both
                `OsmTagsFilter` or `GroupedOsmTagsFilter`.

        Returns:
            osm_tags_type: Merged filters.
        """
        if is_expected_type(tags, OsmTagsFilter):
            return cast(OsmTagsFilter, tags)
        elif is_expected_type(tags, GroupedOsmTagsFilter):
            return merge_grouped_osm_tags_filter(cast(GroupedOsmTagsFilter, tags))

        raise AttributeError(
            "Provided tags don't match required type definitions"
            " (OsmTagsFilter or GroupedOsmTagsFilter)."
        )

    def _parse_features_gdf_to_groups(
        self, features_gdf: gpd.GeoDataFrame, tags: Union[OsmTagsFilter, GroupedOsmTagsFilter]
    ) -> gpd.GeoDataFrame:
        """
        Optionally group raw OSM features into groups defined in `GroupedOsmTagsFilter`.

        Args:
            features_gdf (gpd.GeoDataFrame): Generated features from the loader.
            tags (Union[OsmTagsFilter, GroupedOsmTagsFilter]): OSM tags filter definition.

        Returns:
            gpd.GeoDataFrame: Parsed features_gdf.
        """
        if is_expected_type(tags, GroupedOsmTagsFilter):
            features_gdf = self._group_features_gdf(features_gdf, cast(GroupedOsmTagsFilter, tags))
        return features_gdf

    def _group_features_gdf(
        self, features_gdf: gpd.GeoDataFrame, group_filter: GroupedOsmTagsFilter
    ) -> gpd.GeoDataFrame:
        """
        Group raw OSM features into groups defined in `GroupedOsmTagsFilter`.

        Creates new features based on definition from `GroupedOsmTagsFilter`.
        Returns transformed GeoDataFrame with columns based on group names from the filter.
        Values are built by concatenation of matching tag key and value with
        an equal sign (eg. amenity=parking). Since many tags can match a definition
        of a single group, a first match is used as a feature value.

        Args:
            features_gdf (gpd.GeoDataFrame): Generated features from the loader.
            group_filter (GroupedOsmTagsFilter): Grouped OSM tags filter definition.

        Returns:
            gpd.GeoDataFrame: Parsed grouped features_gdf.
        """
        if len(features_gdf) == 0:
            return features_gdf[["geometry"]]

        matching_columns = []

        for group_name, osm_filter in tqdm(
            group_filter.items(), desc="Grouping features", total=len(group_filter)
        ):
            mask = self._get_matching_mask(features_gdf, osm_filter)
            if mask.any():
                group_name_column = f"{OSMLoader.OSM_FILTER_GROUP_COLUMN_NAME}{group_name}"
                matching_columns.append(group_name_column)
                features_gdf[group_name_column] = features_gdf[mask].apply(
                    lambda row, osm_filter=osm_filter: self._get_first_matching_osm_tag_value(
                        row=row, osm_filter=osm_filter
                    ),
                    axis=1,
                )

        return (
            features_gdf[["geometry", *matching_columns]]
            .rename(
                columns={
                    column_name: column_name.replace(OSMLoader.OSM_FILTER_GROUP_COLUMN_NAME, "")
                    for column_name in matching_columns
                }
            )
            .replace(to_replace=[None], value=np.nan)
            .dropna(how="all", axis="columns")
        )

    def _get_matching_mask(
        self, features_gdf: gpd.GeoDataFrame, osm_filter: OsmTagsFilter
    ) -> pd.Series:
        """
        Create a boolean mask to identify rows matching the OSM tags filter.

        Args:
            features_gdf (gpd.GeoDataFrame): Generated features from the loader.
            osm_filter (OsmTagsFilter): OSM tags filter definition.

        Returns:
            pd.Series: Boolean mask.
        """
        mask = pd.Series(False, index=features_gdf.index)

        for osm_tag_key, osm_tag_value in osm_filter.items():
            if osm_tag_key in features_gdf.columns:
                if isinstance(osm_tag_value, bool) and osm_tag_value:
                    mask |= features_gdf[osm_tag_key]
                elif isinstance(osm_tag_value, str):
                    mask |= features_gdf[osm_tag_key] == osm_tag_value
                elif isinstance(osm_tag_value, list):
                    mask |= features_gdf[osm_tag_key].isin(osm_tag_value)

        return mask

    def _get_first_matching_osm_tag_value(
        self, row: pd.Series, osm_filter: OsmTagsFilter
    ) -> Optional[str]:
        """
        Find first matching OSM tag key and value pair for a subgroup filter.

        Returns a first matching pair of OSM tag key and value concatenated
        with an equal sign (eg. amenity=parking). If none of the values
        in the row matches the filter, `None` value is returned.

        Args:
            row (pd.Series): Row to be analysed.
            osm_filter (osm_tags_type): OSM tags filter definition.

        Returns:
            Optional[str]: New feature value.
        """
        for osm_tag_key, osm_tag_value in osm_filter.items():
            if osm_tag_key not in row or pd.isna(row[osm_tag_key]):
                continue

            is_matching_bool_filter = isinstance(osm_tag_value, bool) and osm_tag_value
            is_matching_string_filter = (
                isinstance(osm_tag_value, str) and row[osm_tag_key] == osm_tag_value
            )
            is_matching_list_filter = (
                isinstance(osm_tag_value, list) and row[osm_tag_key] in osm_tag_value
            )

            if is_matching_bool_filter or is_matching_string_filter or is_matching_list_filter:
                return f"{osm_tag_key}={row[osm_tag_key]}"

        return None



"""
HexagonalDataset.

This dataset is used to train a hexagonal encoder model.
As defined in GeoVex paper[1].

References:
    [1] https://openreview.net/forum?id=7bvWopYY1H
"""

from typing import TYPE_CHECKING, Any, Generic, List, Set, Tuple, TypeVar

import numpy as np
import pandas as pd
from tqdm import tqdm

from srai._optional import import_optional_dependencies
from srai.h3 import get_local_ij_index
from srai.neighbourhoods import H3Neighbourhood

if TYPE_CHECKING:  # pragma: no cover
    import torch

try:  # pragma: no cover
    from torch.utils.data import Dataset

except ImportError:
    from srai.embedders._pytorch_stubs import Dataset


T = TypeVar("T")

# define a type for the dataset item
Ij_Index = Tuple[int, int]
Neighbors = List[Tuple[int, Ij_Index]]
CellInfo = Tuple[str, int, Neighbors]


class HexagonalDataset(Dataset["torch.Tensor"], Generic[T]):  # type: ignore
    """
    Dataset for the hexagonal encoder model.

    It works by returning a 3d tensor of hexagonal regions. The tensor is a cube with the target
    hexagonal region in the center, and the rings of neighbors around surrounding it.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        neighbourhood: H3Neighbourhood,
        neighbor_k_ring: int = 6,
    ):
        """
        Initialize the HexagonalDataset.

        Args:
            data (pd.DataFrame): Data to use for training. Raw counts of features in regions.
            neighbourhood (H3Neighbourhood): H3Neighbourhood to use for training.
                It has to be initialized with the same data as the data argument.
            neighbor_k_ring (int, optional): The hexagonal rings of neighbors to include
                in the tensor. Defaults to 6.
        """
        import_optional_dependencies(dependency_group="torch", modules=["torch"])
        import torch

        self._assert_k_ring_correct(neighbor_k_ring)
        self._assert_h3_neighbourhood(neighbourhood)
        # store the desired k
        self._k: int = neighbor_k_ring
        # number of columns in the dataset
        self._N: int = data.shape[1]
        # store the list of valid h3 indices (have all the neighbors in the dataset)
        self._valid_cells: List[CellInfo] = []
        # store the data as a torch tensor
        self._data_torch = torch.Tensor(data.to_numpy(dtype=np.float32))
        # iterate over the data and build the valid h3 indices
        self._invalid_cells, self._valid_cells = self._seperate_valid_invalid_cells(
            data, neighbourhood, neighbor_k_ring, set(data.index)
        )

    def _seperate_valid_invalid_cells(
        self,
        data: pd.DataFrame,
        neighbourhood: H3Neighbourhood,
        neighbor_k_ring: int,
        all_indices: Set[str],
    ) -> Tuple[Set[str], List[CellInfo]]:
        invalid_h3s = set()
        valid_h3s = []

        for h3_index in tqdm(data.index, total=len(data)):
            neighbors = neighbourhood.get_neighbours_up_to_distance(
                h3_index, neighbor_k_ring, include_center=False, unchecked=True
            )
            # check if all the neighbors are in the dataset
            if len(neighbors.intersection(all_indices)) == len(neighbors):
                # add the h3_index to the valid h3 indices, with the ring of neighbors
                valid_h3s.append(
                    (
                        h3_index,
                        data.index.get_loc(h3_index),
                        [
                            # get the index of the h3 in the dataset
                            (data.index.get_loc(_h), get_local_ij_index(h3_index, _h))
                            for _h in neighbors
                        ],
                    )
                )
            else:
                # some of the neighbors are not in the dataset, add the h3_index to the invalid h3s
                invalid_h3s.add(h3_index)
        return invalid_h3s, valid_h3s

    def __len__(self) -> int:
        """
        Returns the number of valid h3 indices in the dataset.

        Returns:
            int: Number of valid h3 indices in the dataset.
        """
        return len(self._valid_cells)

    def __getitem__(self, index: Any) -> "torch.Tensor":
        """
        Return a single item from the dataset.

        Args:
            index (Any): The index of dataset item to return

        Returns:
            HexagonalDatasetItem: The dataset item
        """
        _, target_idx, neighbors_idxs = self._valid_cells[index]
        return self._build_tensor(target_idx, neighbors_idxs)

    def _build_tensor(self, target_idx: int, neighbors_idxs: Neighbors) -> "torch.Tensor":
        import torch

        # build the 3d tensor
        # it is a tensor with diagonals of length neighbor_k_ring
        # the diagonals are the neighbors of the target h3
        # the target h3 is in the center of the tensor
        # the tensor is 2*neighbor_k_ring + 1 x 2*neighbor_k_ring + 1 x 2*neighbor_k_ring + 1
        # make a tensor of zeros, padded by 1 zero all around to make it even for the convolutions
        tensor: torch.Tensor = torch.zeros(
            (
                self._N,
                2 * self._k + 2,
                2 * self._k + 2,
            )
        )

        # set the target h3 to the center of the tensor
        tensor[
            :,
            self._k,
            self._k,
        ] = self._data_torch[target_idx]

        # set the neighbors of the target h3 to the diagonals of the tensor
        for neighbor_idx, (i, j) in neighbors_idxs:
            tensor[:, self._k + i, self._k - j] = self._data_torch[neighbor_idx]

        # return the tensor and the target (which is same as the tensor)
        # should we return the target as a copy of the tensor?
        return tensor

    def _assert_k_ring_correct(self, k_ring: int) -> None:
        if k_ring < 2:
            raise ValueError(f"k_ring must be at least 2, but was {k_ring}")

    def _assert_h3_neighbourhood(self, neighbourhood: H3Neighbourhood) -> None:
        # force that the neighbourhood is an H3Neighbourhood,
        # because we need the get_ij_index method
        if not isinstance(neighbourhood, H3Neighbourhood):
            raise ValueError(
                f"neighbourhood has to be an H3Neighbourhood, but was {type(neighbourhood)}"
            )

    def get_valid_cells(self) -> List[str]:
        """
        Returns the list of valid h3 indices in the dataset.

        Returns:
            List[str]: List of valid h3 indices in the dataset.
        """
        return [h3_index for h3_index, _, _ in self._valid_cells]

    def get_invalid_cells(self) -> List[str]:
        """
        Returns the list of invalid h3 indices in the dataset.

        Returns:
            List[str]: List of invalid h3 indices in the dataset.
        """
        return list(self._invalid_cells)



"""
H3 neighbourhood.

This module contains the H3Neighbourhood class, that allows to get the neighbours of an H3 region.
"""
from typing import Optional

import geopandas as gpd
import h3

from srai.neighbourhoods import Neighbourhood


class H3Neighbourhood(Neighbourhood[str]):
    """
    H3 Neighbourhood.

    This class allows to get the neighbours of an H3 region.
    """

    def __init__(
        self, regions_gdf: Optional[gpd.GeoDataFrame] = None, include_center: bool = False
    ) -> None:
        """
        Initializes the H3Neighbourhood.

        If a regions GeoDataFrame is provided, only the neighbours
        that are in the regions GeoDataFrame will be returned by the methods of this instance.
        NOTICE: If a region is a part of the k-th ring of a region
            and is included in the GeoDataFrame, it will be returned
            by get_neighbours_at_distance method with distance k
            even when there is no path of length k between the two regions.

        Args:
            regions_gdf (Optional[gpd.GeoDataFrame], optional): The regions that are being analyzed.
                The H3Neighbourhood will only look for neighbours among these regions.
                Defaults to None.
            include_center (bool): Whether to include the region itself in the neighbours.
            This is the default value used for all the methods of the class,
            unless overridden in the function call.
        """
        super().__init__(include_center)
        self._available_indices: Optional[set[str]] = None
        if regions_gdf is not None:
            self._available_indices = set(regions_gdf.index)

    def get_neighbours(self, index: str, include_center: Optional[bool] = None) -> set[str]:
        """
        Get the direct neighbours of an H3 region using its index.

        Args:
            index (str): H3 index of the region.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
            If None, the value set in __init__ is used. Defaults to None.

        Returns:
            Set[str]: Indexes of the neighbours.
        """
        return self.get_neighbours_up_to_distance(index, 1, include_center)

    def get_neighbours_up_to_distance(
        self,
        index: str,
        distance: int,
        include_center: Optional[bool] = None,
        unchecked: bool = False,
    ) -> set[str]:
        """
        Get the neighbours of an H3 region up to a certain distance.

        Args:
            index (str): H3 index of the region.
            distance (int): Distance to the neighbours.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
                If None, the value set in __init__ is used. Defaults to None.
            unchecked (bool): Whether to check if the neighbours are in the available indices.

        Returns:
            Set[str]: Indexes of the neighbours up to the given distance.
        """
        if self._distance_incorrect(distance):
            return set()

        neighbours: set[str] = h3.grid_disk(index, distance)
        neighbours = self._handle_center(
            index, distance, neighbours, at_distance=False, include_center_override=include_center
        )
        if unchecked:
            return neighbours
        return self._select_available(neighbours)

    def get_neighbours_at_distance(
        self, index: str, distance: int, include_center: Optional[bool] = None
    ) -> set[str]:
        """
        Get the neighbours of an H3 region at a certain distance.

        Args:
            index (str): H3 index of the region.
            distance (int): Distance to the neighbours.
            include_center (Optional[bool]): Whether to include the region itself in the neighbours.
            If None, the value set in __init__ is used. Defaults to None.

        Returns:
            Set[str]: Indexes of the neighbours at the given distance.
        """
        if self._distance_incorrect(distance):
            return set()

        neighbours: set[str] = h3.grid_ring(index, distance)
        neighbours = self._handle_center(
            index, distance, neighbours, at_distance=True, include_center_override=include_center
        )
        return self._select_available(neighbours)

    def _select_available(self, indices: set[str]) -> set[str]:
        if self._available_indices is None:
            return indices
        return indices.intersection(self._available_indices)

    def _distance_incorrect(self, distance: int) -> bool:
        return distance < 0



"""H3 shapely conversion tests."""

from typing import Any, Callable, List
from unittest import TestCase

import geopandas as gpd
import pytest
from shapely.geometry.base import BaseGeometry

from srai.constants import GEOMETRY_COLUMN
from srai.h3 import shapely_geometry_to_h3
from tests.regionalizers.test_h3_regionalizer import H3_RESOLUTION

ut = TestCase()


def _gdf_noop(gdf_fixture: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    return gdf_fixture


def _gdf_to_geoseries(gdf_fixture: gpd.GeoDataFrame) -> gpd.GeoSeries:
    return gdf_fixture[GEOMETRY_COLUMN]


def _gdf_to_geometry_list(gdf_fixture: gpd.GeoDataFrame) -> List[BaseGeometry]:
    return list(gdf_fixture[GEOMETRY_COLUMN])


def _gdf_to_single_geometry(gdf_fixture: gpd.GeoDataFrame) -> BaseGeometry:
    return gdf_fixture[GEOMETRY_COLUMN][0]


@pytest.mark.parametrize(
    "geometry_fixture, resolution, expected_h3_cells_fixture",
    [
        ("gdf_single_point", 10, "expected_point_h3_index"),
        ("gdf_multipolygon", H3_RESOLUTION, "expected_unbuffered_h3_indexes"),
        ("gdf_polygons", H3_RESOLUTION, "expected_unbuffered_h3_indexes"),
    ],
)  # type: ignore
@pytest.mark.parametrize(
    "geometry_parser_function",
    [_gdf_noop, _gdf_to_geoseries, _gdf_to_geometry_list],
)  # type: ignore
def test_shapely_geometry_to_h3_unbuffered(
    geometry_fixture: str,
    resolution: int,
    expected_h3_cells_fixture: str,
    geometry_parser_function: Callable[[gpd.GeoDataFrame], Any],
    request: pytest.FixtureRequest,
) -> None:
    """Test checks if conversion from shapely to h3 works."""
    geometry = request.getfixturevalue(geometry_fixture)
    expected_h3_cells = request.getfixturevalue(expected_h3_cells_fixture)

    parsed_geometry = geometry_parser_function(geometry)
    h3_cells = shapely_geometry_to_h3(
        geometry=parsed_geometry, h3_resolution=resolution, buffer=False
    )
    ut.assertCountEqual(h3_cells, expected_h3_cells)


@pytest.mark.parametrize(
    "geometry_fixture, resolution, expected_h3_cells_fixture",
    [
        ("gdf_single_point", 10, "expected_point_h3_index"),
        ("gdf_multipolygon", H3_RESOLUTION, "expected_h3_indexes"),
        ("gdf_polygons", H3_RESOLUTION, "expected_h3_indexes"),
    ],
)  # type: ignore
@pytest.mark.parametrize(
    "geometry_parser_function",
    [_gdf_noop, _gdf_to_geoseries, _gdf_to_geometry_list],
)  # type: ignore
def test_shapely_geometry_to_h3_buffered(
    geometry_fixture: str,
    resolution: int,
    expected_h3_cells_fixture: str,
    geometry_parser_function: Callable[[gpd.GeoDataFrame], Any],
    request: pytest.FixtureRequest,
) -> None:
    """Test checks if conversion from shapely to h3 with buffer works."""
    geometry = request.getfixturevalue(geometry_fixture)
    expected_h3_cells = request.getfixturevalue(expected_h3_cells_fixture)

    parsed_geometry = geometry_parser_function(geometry)
    h3_cells = shapely_geometry_to_h3(
        geometry=parsed_geometry, h3_resolution=resolution, buffer=True
    )
    ut.assertCountEqual(h3_cells, expected_h3_cells)



"""Intersection joiner tests."""
from unittest import TestCase

import geopandas as gpd
import pandas as pd
import pytest

from srai.constants import GEOMETRY_COLUMN
from srai.joiners import IntersectionJoiner

ut = TestCase()


def test_regions_without_geometry_value_error(
    no_geometry_gdf: gpd.GeoDataFrame, features_gdf: gpd.GeoDataFrame
) -> None:
    """Test checks if regions without geometry are disallowed."""
    with pytest.raises(ValueError):
        IntersectionJoiner().transform(regions=no_geometry_gdf, features=features_gdf)


def test_features_without_geometry_value_error(
    regions_gdf: gpd.GeoDataFrame, no_geometry_gdf: gpd.GeoDataFrame
) -> None:
    """Test checks if features without geometry are disallowed."""
    with pytest.raises(ValueError):
        IntersectionJoiner().transform(regions=regions_gdf, features=no_geometry_gdf)


def test_empty_regions_value_error(
    empty_gdf: gpd.GeoDataFrame, features_gdf: gpd.GeoDataFrame
) -> None:
    """Test checks if empty regions are disallowed."""
    with pytest.raises(ValueError):
        IntersectionJoiner().transform(regions=empty_gdf, features=features_gdf)


def test_empty_features_value_error(
    regions_gdf: gpd.GeoDataFrame, empty_gdf: gpd.GeoDataFrame
) -> None:
    """Test checks if empty features are disallowed."""
    with pytest.raises(ValueError):
        IntersectionJoiner().transform(regions=regions_gdf, features=empty_gdf)


def test_correct_multiindex_intersection_joiner(
    regions_gdf: gpd.GeoDataFrame, features_gdf: gpd.GeoDataFrame, joint_multiindex: pd.MultiIndex
) -> None:
    """Test checks if intersection joiner returns correct MultiIndex."""
    joint = IntersectionJoiner().transform(
        regions=regions_gdf, features=features_gdf, return_geom=True
    )

    ut.assertEqual(joint.index.names, joint_multiindex.names)
    ut.assertCountEqual(joint.index, joint_multiindex)
    ut.assertIn(GEOMETRY_COLUMN, joint.columns)


def test_correct_multiindex_intersection_joiner_without_geom(
    regions_gdf: gpd.GeoDataFrame, features_gdf: gpd.GeoDataFrame, joint_multiindex: pd.MultiIndex
) -> None:
    """Test checks if intersection joiner returns correct MultiIndex."""
    joint = IntersectionJoiner().transform(
        regions=regions_gdf, features=features_gdf, return_geom=False
    )

    ut.assertEqual(joint.index.names, joint_multiindex.names)
    ut.assertCountEqual(joint.index, joint_multiindex)
    ut.assertNotIn(GEOMETRY_COLUMN, joint.columns)
    ut.assertIs(len(joint.columns), 0)



"""Base class for loaders."""

import abc
from typing import Any

import geopandas as gpd


class Loader(abc.ABC):
    """Abstract class for loaders."""

    @abc.abstractmethod
    def load(self, *args: Any, **kwargs: Any) -> gpd.GeoDataFrame:  # pragma: no cover
        """
        Load data for a given area.

        Args:
            *args: Positional arguments dependating on a specific loader.
            **kwargs: Keyword arguments dependating on a specific loader.

        Returns:
            GeoDataFrame with the downloaded data.
        """
        raise NotImplementedError



from enum import Enum
from types import ModuleType
from typing import List, Optional


class ImportErrorHandle(Enum):
    """List of values defining how to handle import errors."""

    RAISE = "raise"
    WARN = "warn"
    IGNORE = "ignore"


# Inspired by:
# https://github.com/pandas-dev/pandas/blob/main/pandas/compat/_optional.py
# https://stackoverflow.com/questions/563022/whats-python-good-practice-for-importing-and-offering-optional-features
def import_optional_dependency(
    dependency_group: str,
    module: str,
    name: Optional[str] = None,
    error: ImportErrorHandle = ImportErrorHandle.RAISE,
) -> Optional[ModuleType]:
    """
    Import a module or a element from the module.

    Args:
        dependency_group (str): Name of dependency group where dependency is defined.
            Helps communicate a proper error message to the end user.
        module (str): Name of a module.
        name (str, optional): Name of element from the module. If none, returns whole module.
            Otherwise returns a submodule found using a given name. Defaults to None.
        error (ErrorHandle, {'raise', 'warn', 'ignore'}): Information what to do when module hasn't
            been found. Can `raise` an error, write a warning (`warn`) or `ignore` missing module.
            Defaults to ErrorHandle.RAISE.

    Raises:
        ImportError: When required dependency is not installed.

    Returns:
        Optional[ModuleType]: Module or submodule imported using a name. None if not found.
    """
    assert error in ImportErrorHandle.__members__.values()
    import importlib

    try:
        imported_module = importlib.import_module(module)
        return imported_module if name is None else getattr(imported_module, name)
    except ImportError as exc:
        error_msg = (
            f'Missing optional dependency "{module}". Please install required packages using '
            f"`pip install srai[{dependency_group}]`."
        )
        if error == ImportErrorHandle.RAISE:
            raise ImportError(error_msg) from exc
        if error == ImportErrorHandle.WARN:
            import warnings

            warnings.warn(f"{error_msg} Skipping import.", ImportWarning, stacklevel=2)
    return None


def import_optional_dependencies(
    dependency_group: str, modules: List[str], error: ImportErrorHandle = ImportErrorHandle.RAISE
) -> None:
    """
    Import list of optional dependencies.

    Args:
        dependency_group (str): Name of optional group that contains dependencies.
        modules (List[str]): List of module names that are expected to be imported.
        error (ErrorHandle, {'raise', 'warn', 'ignore'}): Information what to do when any of
            the modules hasn't been found. Defaults to ErrorHandle.RAISE.
    """
    for module in modules:
        import_optional_dependency(dependency_group=dependency_group, module=module, error=error)



"""
Administrative Boundary Regionalizer.

This module contains administrative boundary regionalizer implementation.
"""
import hashlib
import json
import time
from pathlib import Path
from typing import Any, Optional, Union

import geopandas as gpd
import topojson as tp
from shapely import union
from shapely.geometry import GeometryCollection, MultiPolygon, Point, Polygon
from shapely.geometry.base import BaseGeometry
from shapely.validation import make_valid
from tqdm import tqdm

from srai._optional import import_optional_dependencies
from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.geometry import flatten_geometry_series
from srai.regionalizers import Regionalizer


class AdministrativeBoundaryRegionalizer(Regionalizer):
    """
    AdministrativeBoundaryRegionalizer.

    Administrative boundary regionalizer allows the given geometries to be divided
    into boundaries from OpenStreetMap on a given `admin_level` [1].

    Downloads those boundaries online using `overpass` and `osmnx` libraries.

    Note: offline .pbf loading will be implemented in the future.
    Note: option to download historic data will be implemented in the future.

    References:
        1. https://wiki.openstreetmap.org/wiki/Key:admin_level
    """

    EMPTY_REGION_NAME = "EMPTY"

    def __init__(
        self,
        admin_level: int,
        clip_regions: bool = True,
        return_empty_region: bool = False,
        prioritize_english_name: bool = True,
        toposimplify: Union[bool, float] = True,
        remove_artefact_regions: bool = True,
    ) -> None:
        """
        Init AdministrativeBoundaryRegionalizer.

        Args:
            admin_level (int): OpenStreetMap admin_level value. See [1] for detailed description of
                available values.
            clip_regions (bool, optional): Whether to clip regions using a provided mask.
                Turning it off can an be useful when trying to load regions using list a of points.
                Defaults to True.
            return_empty_region (bool, optional): Whether to return an empty region to fill
                remaining space or not. Defaults to False.
            prioritize_english_name (bool, optional): Whether to use english area name if available
                as a region id first. Defaults to True.
            toposimplify (Union[bool, float], optional): Whether to simplify topology to reduce
                geometries size or not. Value is passed to `topojson` library for topology-aware
                simplification. Since provided values are treated like degrees, values between
                1e-4 and 1.0 are recommended. Defaults to True (which results in value equal 1e-4).
            remove_artefact_regions (bool, optional): Whether to remove small regions barely
                intersecting queried area. Turning it off can sometimes load unnecessary boundaries
                that touch on the edge. It removes regions that intersect with an area smaller
                than 1% of total self. Takes into consideration if provided query GeoDataFrame
                contains points and then skips calculating area when intersects any point.
                Defaults to True.

        Raises:
            ValueError: If admin_level is outside available range (1-11). See [2] for list of
                values with `in_wiki` selected.

        References:
            1. https://wiki.openstreetmap.org/wiki/Tag:boundary=administrative#10_admin_level_values_for_specific_countries
            2. https://taginfo.openstreetmap.org/keys/admin_level#values
        """  # noqa: W505, E501
        import_optional_dependencies(
            dependency_group="osm",
            modules=["osmnx", "overpass"],
        )
        from overpass import API

        if admin_level < 1 or admin_level > 11:
            raise ValueError("admin_level outside of available range.")

        self.admin_level = admin_level
        self.prioritize_english_name = prioritize_english_name
        self.clip_regions = clip_regions
        self.return_empty_region = return_empty_region
        self.remove_artefact_regions = remove_artefact_regions

        if isinstance(toposimplify, (int, float)) and toposimplify > 0:
            self.toposimplify = toposimplify
        elif isinstance(toposimplify, bool) and toposimplify:
            self.toposimplify = 1e-4
        else:
            self.toposimplify = False

        self.overpass_api = API(timeout=60)

    def transform(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Regionalize a given GeoDataFrame.

        Will query Overpass [1] server using `overpass` [2] library for closed administrative
        boundaries on a given admin_level and then download geometries for each relation using
        `osmnx` [3] library.

        If `prioritize_english_name` is set to `True`, method will try to extract the `name:en` tag
        first before resorting to the `name` tag. If boundary doesn't have a `name` tag, an `id`
        will be used.

        Before returning downloaded regions, a topology is built and can be simplified to reduce
        size of geometries while keeping neighbouring regions together without introducing gaps.
        `Topojson` library [4] is used for this operation.

        Additionally, an empty region with name `EMPTY` can be introduced if returned regions do
        not fully cover a clipping mask.

        Args:
            gdf (gpd.GeoDataFrame): GeoDataFrame to be regionalized.
                Will use this list of geometries to crop resulting regions.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with the regionalized data cropped using input
                GeoDataFrame.

        Raises:
            RuntimeError: If simplification can't preserve a topology.

        References:
            1. https://wiki.openstreetmap.org/wiki/Overpass_API
            2. https://github.com/mvexel/overpass-api-python-wrapper
            3. https://github.com/gboeing/osmnx
            4. https://github.com/mattijn/topojson
        """
        gdf_wgs84 = gdf.to_crs(crs=WGS84_CRS)

        regions_dicts = self._generate_regions_from_all_geometries(gdf_wgs84)

        if not regions_dicts:
            import warnings

            warnings.warn(
                "Couldn't find any administrative boundaries with"
                f" `admin_level`={self.admin_level}.",
                RuntimeWarning,
                stacklevel=2,
            )

            return self._get_empty_geodataframe(gdf_wgs84)

        regions_gdf = gpd.GeoDataFrame(data=regions_dicts, crs=WGS84_CRS).set_index(REGIONS_INDEX)
        regions_gdf = self._toposimplify_gdf(regions_gdf)

        if self.remove_artefact_regions:
            points_collection: Optional[BaseGeometry] = gdf_wgs84[
                gdf_wgs84.geom_type == "Point"
            ].geometry.unary_union
            clipping_polygon_area: Optional[BaseGeometry] = gdf_wgs84[
                gdf_wgs84.geom_type != "Point"
            ].geometry.unary_union

            regions_to_keep = [
                region_id
                for region_id, row in regions_gdf.iterrows()
                if self._check_intersects_with_points(row["geometry"], points_collection)
                or self._calculate_intersection_area_fraction(
                    row["geometry"], clipping_polygon_area
                )
                > 0.01
            ]
            regions_gdf = regions_gdf.loc[regions_to_keep]

        if self.clip_regions:
            regions_gdf = regions_gdf.clip(mask=gdf_wgs84, keep_geom_type=False)

        if self.return_empty_region:
            empty_region = self._generate_empty_region(mask=gdf_wgs84, regions_gdf=regions_gdf)
            if not empty_region.is_empty:
                regions_gdf.loc[
                    AdministrativeBoundaryRegionalizer.EMPTY_REGION_NAME, GEOMETRY_COLUMN
                ] = empty_region

        return regions_gdf

    def _generate_regions_from_all_geometries(
        self, gdf_wgs84: gpd.GeoDataFrame
    ) -> list[dict[str, Any]]:
        """Query and optimize downloading data from Overpass."""
        elements_ids = set()
        generated_regions: list[dict[str, Any]] = []
        unary_geometry = GeometryCollection()

        all_geometries = flatten_geometry_series(gdf_wgs84.geometry)

        with tqdm(desc="Loading boundaries: 0", total=len(all_geometries)) as pbar:
            for geometry in all_geometries:
                if not geometry.covered_by(unary_geometry):
                    query = self._generate_query_for_single_geometry(geometry)
                    boundaries_list = self._query_overpass(query)
                    for boundary in boundaries_list:
                        if boundary["id"] not in elements_ids:
                            elements_ids.add(boundary["id"])
                            parsed_region = self._parse_overpass_element(boundary)
                            generated_regions.append(parsed_region)
                            unary_geometry = union(unary_geometry, parsed_region[GEOMETRY_COLUMN])
                            pbar.set_description(f"Loading boundaries: {len(generated_regions)}")
                pbar.update(1)

        return generated_regions

    def _query_overpass(self, query: str) -> list[dict[str, Any]]:
        """
        Query Overpass and catch exceptions.

        This method allows for retry after waiting some time, and caches the response.

        Args:
            query (str): Overpass query.

        Raises:
            ex: If exception is different than urllib.request.HTTPError or
                HTTP code is different than 429 or 504.

        Returns:
            List[Dict[str, Any]]: Query elements result from Overpass.
        """
        from overpass import MultipleRequestsError, ServerLoadError

        h = hashlib.new("sha256")
        h.update(query.encode())
        query_hash = h.hexdigest()
        query_file_path = Path("cache").resolve() / f"{query_hash}.json"

        while True:
            try:
                if not query_file_path.exists():
                    query_result = self.overpass_api.get(
                        query, verbosity="ids tags", responseformat="json"
                    )
                    query_file_path.parent.mkdir(parents=True, exist_ok=True)
                    query_file_path.write_text(json.dumps(query_result))
                else:
                    query_result = json.loads(query_file_path.read_text())

                elements: list[dict[str, Any]] = query_result["elements"]
                return elements
            except (MultipleRequestsError, ServerLoadError):
                time.sleep(60)

    def _generate_query_for_single_geometry(self, g: BaseGeometry) -> str:
        """Generate Overpass query for a geometry."""
        if isinstance(g, Point):
            query = (
                f"is_in({g.y},{g.x});"
                'area._["boundary"="administrative"]'
                '["type"~"boundary|multipolygon"]'
                f'["admin_level"="{self.admin_level}"];'
                "relation(pivot);"
            )
        else:
            raw_gdf_bounds = g.bounds
            overpass_bbox = (
                raw_gdf_bounds[1],
                raw_gdf_bounds[0],
                raw_gdf_bounds[3],
                raw_gdf_bounds[2],
            )
            query = (
                '(relation["boundary"="administrative"]'
                '["type"~"boundary|multipolygon"]'
                f'["admin_level"="{self.admin_level}"]'
                f"({overpass_bbox[0]},{overpass_bbox[1]},{overpass_bbox[2]},{overpass_bbox[3]}););"
            )
        return query

    def _parse_overpass_element(self, element: dict[str, Any]) -> dict[str, Any]:
        """Parse single Overpass Element and return a region."""
        element_tags = element.get("tags", {})
        region_id = None
        if self.prioritize_english_name:
            region_id = element_tags.get("name:en")
        if not region_id:
            region_id = element_tags.get("name")
        if not region_id:
            region_id = str(element["id"])

        return {
            GEOMETRY_COLUMN: self._get_boundary_geometry(element["id"]),
            REGIONS_INDEX: region_id,
        }

    def _get_boundary_geometry(self, relation_id: str) -> BaseGeometry:
        """Download a geometry of a relation using `osmnx` library."""
        from osmnx import geocode_to_gdf

        return geocode_to_gdf(query=[f"R{relation_id}"], by_osmid=True).geometry[0]

    def _toposimplify_gdf(self, regions_gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Create a topology to ensure proper boundaries between regions and simplify it."""
        topo = tp.Topology(
            regions_gdf,
            prequantize=False,
            presimplify=False,
            toposimplify=self.toposimplify,
            simplify_algorithm="dp",
            prevent_oversimplify=True,
        )
        regions_gdf = topo.to_gdf(winding_order="CW_CCW", crs=WGS84_CRS, validate=True)
        regions_gdf.index.rename(REGIONS_INDEX, inplace=True)
        regions_gdf.geometry = regions_gdf.geometry.apply(make_valid)
        for idx, r in regions_gdf.iterrows():
            if not isinstance(r.geometry, (Polygon, MultiPolygon)):
                raise RuntimeError(
                    f"Simplification couldn't preserve geometry for region: {idx}."
                    " Try lowering toposimplify value."
                )
        return regions_gdf

    def _generate_empty_region(
        self, mask: gpd.GeoDataFrame, regions_gdf: gpd.GeoDataFrame
    ) -> BaseGeometry:
        """Generate a region filling the space between regions and full clipping mask."""
        joined_mask = mask.geometry.unary_union
        joined_geometry = regions_gdf.geometry.unary_union
        return joined_mask.difference(joined_geometry)

    def _get_empty_geodataframe(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """Get empty GeoDataFrame when zero boundaries have been found."""
        if self.return_empty_region:
            regions_gdf = gpd.GeoDataFrame(
                data={
                    GEOMETRY_COLUMN: [gdf.geometry.unary_union],
                    REGIONS_INDEX: [AdministrativeBoundaryRegionalizer.EMPTY_REGION_NAME],
                },
                crs=WGS84_CRS,
            ).set_index(REGIONS_INDEX)
        else:
            regions_gdf = gpd.GeoDataFrame(
                data={
                    GEOMETRY_COLUMN: [],
                    REGIONS_INDEX: [],
                },
                crs=WGS84_CRS,
            )
        return regions_gdf

    def _check_intersects_with_points(
        self, region_geometry: BaseGeometry, points_collection: Optional[BaseGeometry]
    ) -> bool:
        """Check if region intersects with any point in query regions."""
        return (
            points_collection is not None
            and not points_collection.is_empty
            and region_geometry.intersects(points_collection)
        )

    def _calculate_intersection_area_fraction(
        self, region_geometry: BaseGeometry, clipping_polygon_area: Optional[BaseGeometry]
    ) -> float:
        """Calculate intersection area fraction to check if it's big enough."""
        if clipping_polygon_area is None or clipping_polygon_area.is_empty:
            return 0
        full_area = float(region_geometry.area)
        clip_area = float(region_geometry.intersection(clipping_polygon_area).area)
        return clip_area / full_area



"""Voronoi regionalizer tests."""
from contextlib import nullcontext as does_not_raise
from typing import Any, Union

import geopandas as gpd
import pytest
from overpass import API
from pytest_mock import MockerFixture
from shapely.geometry import Point, box

from srai.constants import GEOMETRY_COLUMN, WGS84_CRS
from srai.geometry import merge_disjointed_gdf_geometries
from srai.regionalizers import AdministrativeBoundaryRegionalizer

bbox = box(minx=-180, maxx=180, miny=-90, maxy=90)
bbox_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [bbox]}, crs=WGS84_CRS)


@pytest.mark.parametrize(  # type: ignore
    "admin_level,expectation",
    [
        (-1, pytest.raises(ValueError)),
        (0, pytest.raises(ValueError)),
        (1, does_not_raise()),
        (2, does_not_raise()),
        (3, does_not_raise()),
        (4, does_not_raise()),
        (5, does_not_raise()),
        (6, does_not_raise()),
        (7, does_not_raise()),
        (8, does_not_raise()),
        (9, does_not_raise()),
        (10, does_not_raise()),
        (11, does_not_raise()),
        (12, pytest.raises(ValueError)),
    ],
)
def test_admin_level(
    admin_level: int,
    expectation: Any,
) -> None:
    """Test checks if illegal admin_level is disallowed."""
    with expectation:
        AdministrativeBoundaryRegionalizer(admin_level=admin_level)


def test_empty_gdf_attribute_error(gdf_empty) -> None:  # type: ignore
    """Test checks if empty GeoDataFrames are disallowed."""
    with pytest.raises(AttributeError):
        abr = AdministrativeBoundaryRegionalizer(admin_level=4)
        abr.transform(gdf_empty)


def test_no_crs_gdf_value_error(gdf_no_crs) -> None:  # type: ignore
    """Test checks if GeoDataFrames without crs are disallowed."""
    with pytest.raises(ValueError):
        abr = AdministrativeBoundaryRegionalizer(admin_level=4)
        abr.transform(gdf=gdf_no_crs)


@pytest.fixture  # type: ignore
def mock_overpass_api(mocker: MockerFixture) -> None:
    """Mock overpass API."""
    mocker.patch.object(API, "get", return_value={"elements": [{"type": "relation", "id": 2137}]})

    geocoded_gdf = gpd.GeoDataFrame(
        {GEOMETRY_COLUMN: [box(minx=0, miny=0, maxx=1, maxy=1)]}, crs=WGS84_CRS
    )
    mocker.patch("osmnx.geocode_to_gdf", return_value=geocoded_gdf)


@pytest.mark.parametrize(  # type: ignore
    "toposimplify",
    [
        (True),
        (0.0001),
        (0.001),
        (0.01),
        (False),
        (0),
    ],
)
def test_empty_region_full_bounding_box(toposimplify: Union[bool, float], request: Any) -> None:
    """Test checks if empty region fills required bounding box."""
    request.getfixturevalue("mock_overpass_api")
    request_bbox = box(minx=0, miny=0, maxx=2, maxy=2)
    request_bbox_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [request_bbox]}, crs=WGS84_CRS)
    abr = AdministrativeBoundaryRegionalizer(
        admin_level=4, return_empty_region=True, toposimplify=toposimplify
    )
    result_gdf = abr.transform(gdf=request_bbox_gdf)
    assert merge_disjointed_gdf_geometries(result_gdf).difference(request_bbox).is_empty
    assert "EMPTY" in result_gdf.index


@pytest.mark.parametrize(  # type: ignore
    "toposimplify",
    [
        (True),
        (0.0001),
        (0.001),
        (0.01),
        (False),
        (0),
    ],
)
def test_no_empty_region_full_bounding_box(toposimplify: Union[bool, float], request: Any) -> None:
    """Test checks if no empty region is generated when not needed."""
    request.getfixturevalue("mock_overpass_api")
    request_bbox = box(minx=0, miny=0, maxx=1, maxy=1)
    request_bbox_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [request_bbox]}, crs=WGS84_CRS)
    abr = AdministrativeBoundaryRegionalizer(
        admin_level=2, return_empty_region=True, toposimplify=toposimplify
    )
    result_gdf = abr.transform(gdf=request_bbox_gdf)
    assert merge_disjointed_gdf_geometries(result_gdf).difference(request_bbox).is_empty
    assert "EMPTY" not in result_gdf.index


@pytest.mark.parametrize(  # type: ignore
    "toposimplify",
    [
        (True),
        (0.0001),
        (0.001),
        (0.01),
        (False),
        (0),
    ],
)
def test_points_in_result(toposimplify: Union[bool, float], request: Any) -> None:
    """Test checks case when points are in a requested region."""
    request.getfixturevalue("mock_overpass_api")
    request_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [Point(0.5, 0.5)]}, crs=WGS84_CRS)

    abr = AdministrativeBoundaryRegionalizer(
        admin_level=2, return_empty_region=False, clip_regions=False, toposimplify=toposimplify
    )

    result_gdf = abr.transform(gdf=request_gdf)
    assert request_gdf.geometry[0].covered_by(result_gdf.geometry[0])


@pytest.mark.parametrize(  # type: ignore
    "toposimplify",
    [
        (True),
        (0.0001),
        (0.001),
        (0.01),
        (False),
        (0),
    ],
)
def test_toposimplify_on_real_data(toposimplify: Union[float, bool]) -> None:
    """Test if toposimplify usage covers an entire region."""
    madagascar_bbox = box(
        minx=43.2541870461, miny=-25.6014344215, maxx=50.4765368996, maxy=-12.0405567359
    )
    madagascar_bbox_gdf = gpd.GeoDataFrame({GEOMETRY_COLUMN: [madagascar_bbox]}, crs=WGS84_CRS)

    abr = AdministrativeBoundaryRegionalizer(
        admin_level=4, return_empty_region=True, toposimplify=toposimplify
    )
    madagascar_result_gdf = abr.transform(gdf=madagascar_bbox_gdf)
    assert (
        merge_disjointed_gdf_geometries(madagascar_result_gdf).difference(madagascar_bbox).is_empty
    )


@pytest.mark.parametrize(  # type: ignore
    "return_empty_region",
    [
        (True),
        (False),
    ],
)
def test_regions_not_found_on_real_data(return_empty_region: bool) -> None:
    """Test if warns when can't find any regions."""
    null_island_region = box(minx=0, miny=0, maxx=0.1, maxy=0.1)
    null_island_region_gdf = gpd.GeoDataFrame(
        {GEOMETRY_COLUMN: [null_island_region]}, crs=WGS84_CRS
    )

    abr = AdministrativeBoundaryRegionalizer(
        admin_level=10, return_empty_region=return_empty_region
    )

    with pytest.warns(RuntimeWarning):
        madagascar_result_gdf = abr.transform(gdf=null_island_region_gdf)

    if return_empty_region:
        assert (
            merge_disjointed_gdf_geometries(madagascar_result_gdf)
            .difference(null_island_region)
            .is_empty
        )
    else:
        assert merge_disjointed_gdf_geometries(madagascar_result_gdf).is_empty



"""
Hex2Vec Embedder.

This module contains embedder from Hex2Vec paper[1].

References:
    [1] https://dl.acm.org/doi/10.1145/3486635.3491076
"""
import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, TypeVar, Union

import geopandas as gpd
import numpy as np
import pandas as pd

from srai._optional import import_optional_dependencies
from srai.embedders import CountEmbedder, ModelT
from srai.embedders.hex2vec.model import Hex2VecModel
from srai.embedders.hex2vec.neighbour_dataset import NeighbourDataset
from srai.exceptions import ModelNotFitException
from srai.loaders.osm_loaders.filters import GroupedOsmTagsFilter, OsmTagsFilter
from srai.neighbourhoods import Neighbourhood

T = TypeVar("T")


class Hex2VecEmbedder(CountEmbedder):
    """Hex2Vec Embedder."""

    DEFAULT_ENCODER_SIZES = [150, 75, 50]

    def __init__(
        self,
        encoder_sizes: Optional[List[int]] = None,
        expected_output_features: Optional[
            Union[List[str], OsmTagsFilter, GroupedOsmTagsFilter]
        ] = None,
    ) -> None:
        """
        Initialize Hex2VecEmbedder.

        Args:
            encoder_sizes (List[int], optional): Sizes of the encoder layers.
                The input layer size shouldn't be included - it's inferred from the data.
                The last element is the embedding size. Defaults to [150, 75, 50].
            expected_output_features
                (Union[List[str], OsmTagsFilter, GroupedOsmTagsFilter], optional):
                List of expected output features. Defaults to None.
        """
        super().__init__(
            expected_output_features=expected_output_features, count_subcategories=True
        )
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )
        if encoder_sizes is None:
            encoder_sizes = Hex2VecEmbedder.DEFAULT_ENCODER_SIZES
        self._assert_encoder_sizes_correct(encoder_sizes)
        self._encoder_sizes = encoder_sizes
        self._model: Optional[Hex2VecModel] = None
        self._is_fitted = False

    def transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
    ) -> pd.DataFrame:
        """
        Create region embeddings.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.

        Returns:
            pd.DataFrame: Embedding and geometry index for each region in regions_gdf.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
        """
        import torch

        self._check_is_fitted()
        counts_df = self._get_raw_counts(regions_gdf, features_gdf, joint_gdf)
        counts_tensor = torch.from_numpy(counts_df.values)
        embeddings = self._model(counts_tensor).detach().numpy()  # type: ignore
        return pd.DataFrame(embeddings, index=counts_df.index)

    def fit(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        neighbourhood: Neighbourhood[T],
        negative_sample_k_distance: int = 2,
        batch_size: int = 32,
        learning_rate: float = 0.001,
        trainer_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Fit the model to the data.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            neighbourhood (Neighbourhood[T]): The neighbourhood to use.
                Should be intialized with the same regions.
            negative_sample_k_distance (int, optional): When sampling negative samples,
                sample from a distance > k. Defaults to 2.
            batch_size (int, optional): Batch size. Defaults to 32.
            learning_rate (float, optional): Learning rate. Defaults to 0.001.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs. Defaults to None.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
            ValueError: If negative_sample_k_distance < 2.
        """
        import pytorch_lightning as pl
        from torch.utils.data import DataLoader

        trainer_kwargs = self._prepare_trainer_kwargs(trainer_kwargs)

        counts_df = self._get_raw_counts(regions_gdf, features_gdf, joint_gdf)

        if self.expected_output_features is None:  # type: ignore[has-type]
            self.expected_output_features = pd.Series(counts_df.columns)

        num_features = len(self.expected_output_features)  # type: ignore[arg-type]
        self._model = Hex2VecModel(
            layer_sizes=[num_features, *self._encoder_sizes], learning_rate=learning_rate
        )
        dataset = NeighbourDataset(counts_df, neighbourhood, negative_sample_k_distance)
        dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True)

        trainer = pl.Trainer(**trainer_kwargs)
        trainer.fit(self._model, dataloader)
        self._is_fitted = True

    def fit_transform(
        self,
        regions_gdf: gpd.GeoDataFrame,
        features_gdf: gpd.GeoDataFrame,
        joint_gdf: gpd.GeoDataFrame,
        neighbourhood: Neighbourhood[T],
        negative_sample_k_distance: int = 2,
        batch_size: int = 32,
        learning_rate: float = 0.001,
        trainer_kwargs: Optional[Dict[str, Any]] = None,
    ) -> pd.DataFrame:
        """
        Fit the model to the data and return the embeddings.

        Args:
            regions_gdf (gpd.GeoDataFrame): Region indexes and geometries.
            features_gdf (gpd.GeoDataFrame): Feature indexes, geometries and feature values.
            joint_gdf (gpd.GeoDataFrame): Joiner result with region-feature multi-index.
            neighbourhood (Neighbourhood[T]): The neighbourhood to use.
                Should be intialized with the same regions.
            negative_sample_k_distance (int, optional): When sampling negative samples,
                sample from a distance > k. Defaults to 2.
            batch_size (int, optional): Batch size. Defaults to 32.
            learning_rate (float, optional): Learning rate. Defaults to 0.001.
            trainer_kwargs (Optional[Dict[str, Any]], optional): Trainer kwargs. Defaults to None.

        Returns:
            pd.DataFrame: Region embeddings.

        Raises:
            ValueError: If features_gdf is empty and self.expected_output_features is not set.
            ValueError: If any of the gdfs index names is None.
            ValueError: If joint_gdf.index is not of type pd.MultiIndex or doesn't have 2 levels.
            ValueError: If index levels in gdfs don't overlap correctly.
            ValueError: If negative_sample_k_distance < 2.
        """
        self.fit(
            regions_gdf,
            features_gdf,
            joint_gdf,
            neighbourhood,
            negative_sample_k_distance,
            batch_size,
            learning_rate,
            trainer_kwargs,
        )
        return self.transform(regions_gdf, features_gdf, joint_gdf)

    def _prepare_trainer_kwargs(self, trainer_kwargs: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        if trainer_kwargs is None:
            trainer_kwargs = {}
        if "max_epochs" not in trainer_kwargs:
            trainer_kwargs["max_epochs"] = 3
        return trainer_kwargs

    def _get_raw_counts(
        self, regions_gdf: pd.DataFrame, features_gdf: pd.DataFrame, joint_gdf: pd.DataFrame
    ) -> pd.DataFrame:
        return super().transform(regions_gdf, features_gdf, joint_gdf).astype(np.float32)

    def _check_is_fitted(self) -> None:
        if not self._is_fitted or self._model is None:
            raise ModelNotFitException("Model not fitted. Call fit() or fit_transform() first.")

    def _assert_encoder_sizes_correct(self, encoder_sizes: List[int]) -> None:
        if len(encoder_sizes) < 1:
            raise ValueError("Encoder sizes must have at least one element - embedding size.")
        if any(size <= 0 for size in encoder_sizes):
            raise ValueError("Encoder sizes must be positive integers.")

    def _save(self, path: Union[Path, str], embedder_config: Dict[str, Any]) -> None:
        if isinstance(path, str):
            path = Path(path)

        self._check_is_fitted()

        path.mkdir(parents=True, exist_ok=True)

        self._model.save(path / "model.pt")  # type: ignore

        config = {
            "model_config": self._model.get_config(),  # type: ignore
            "embedder_config": embedder_config,
        }
        with open(path / "config.json", "w") as f:
            json.dump(config, f, ensure_ascii=False, indent=4)

    def save(self, path: Union[Path, str]) -> None:
        """
        Save the model to a directory.

        Args:
            path (Path): Path to the directory.
        """
        embedder_config = {
            "encoder_sizes": self._encoder_sizes,
            "expected_output_features": (
                self.expected_output_features.tolist()
                if self.expected_output_features is not None
                else None
            ),
        }
        self._save(path, embedder_config)

    @classmethod
    def _load(cls, path: Union[Path, str], model_module: Type[ModelT]) -> "Hex2VecEmbedder":
        if isinstance(path, str):
            path = Path(path)

        with (path / "config.json").open("r") as f:
            config = json.load(f)
        embedder = cls(**config["embedder_config"])
        model_path = path / "model.pt"
        model = model_module.load(model_path, **config["model_config"])
        embedder._model = model
        embedder._is_fitted = True
        return embedder

    @classmethod
    def load(cls, path: Union[Path, str]) -> "Hex2VecEmbedder":
        """
        Load the model from a directory.

        Args:
            path (Path): Path to the directory.

        Returns:
            Hex2VecEmbedder: The loaded embedder.
        """
        return cls._load(path, Hex2VecModel)



from typing import Generic, TypeVar

T_co = TypeVar("T_co", covariant=True)


class Dataset(Generic[T_co]):
    """Dataset class stub."""

    ...


class DataLoader(Generic[T_co]):
    """DataLoader class stub."""

    ...


class LightningModule:  # pragma: no cover
    """LightningModule class stub."""

    ...


class nn:  # pragma: no cover
    """Pytorch nn class stub."""

    class Module:
        """Pytorch nn.Module class stub."""

        ...


class torch:  # pragma: no cover
    """Pytorch class stub."""

    ...



"""
Slippy map regionalizer.

This module implements Slippy map tilenames [1] as a regionalizer.

References:
    1. https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames
"""
from itertools import product
from typing import Any

import geopandas as gpd
import numpy as np
import shapely.geometry as shpg
from functional import seq

from srai.constants import GEOMETRY_COLUMN, REGIONS_INDEX, WGS84_CRS
from srai.regionalizers import Regionalizer


class SlippyMapRegionalizer(Regionalizer):
    """SlippyMapRegionalizer class."""

    def __init__(self, zoom: int) -> None:
        """
        Initialize SlippyMapRegionalizer.

        Args:
            zoom (int): zoom level

        Raises:
            ValueError: if zoom is not in [0, 19]
        """
        if not 0 <= zoom <= 19:
            raise ValueError
        self.zoom = zoom
        super().__init__()

    def transform(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
        """
        Regionalize a given GeoDataFrame.

        Args:
            gdf (gpd.GeoDataFrame): GeoDataFrame to be regionalized.

        Returns:
            gpd.GeoDataFrame: GeoDataFrame with regionalized geometries.

        Raises:
            ValueError: If provided GeoDataFrame has no crs defined.
        """
        gdf_wgs84 = gdf.to_crs(crs=WGS84_CRS)
        gdf_exploded = self._explode_multipolygons(gdf_wgs84)

        values = (
            seq(gdf_exploded[GEOMETRY_COLUMN])
            .map(self._to_cells)
            .flatten()
            .map(
                lambda item: (
                    item
                    | {
                        REGIONS_INDEX: f"{item['x']}_{item['y']}_{self.zoom}",
                        "z": self.zoom,
                    }
                )
            )
            .to_list()
        )

        gdf = gpd.GeoDataFrame(values, geometry=GEOMETRY_COLUMN, crs=WGS84_CRS).set_index(
            REGIONS_INDEX
        )
        return gdf.drop_duplicates()

    def _to_cells(self, polygon: shpg.Polygon) -> list[dict[str, Any]]:
        gdf_bounds = polygon.bounds
        x_start, y_start = self._coordinates_to_x_y(gdf_bounds[1], gdf_bounds[0])
        x_end, y_end = self._coordinates_to_x_y(gdf_bounds[3], gdf_bounds[2])
        tiles = []
        for x, y in product(range(x_start, x_end + 1), range(y_end, y_start + 1)):
            tile_polygon = self._polygon_from_x_y(x, y)
            if not self._should_skip_tile(polygon, tile_polygon):
                tiles.append(dict(x=x, y=y, geometry=tile_polygon))
        return tiles

    def _polygon_from_x_y(self, x: int, y: int) -> shpg.Polygon:
        latitude_start, longitude_start = self._x_y_to_coordinates(x, y)
        latitude_end, longitude_end = self._x_y_to_coordinates(x + 1, y + 1)
        tile_polygon = shpg.box(
            minx=longitude_start, miny=latitude_start, maxx=longitude_end, maxy=latitude_end
        )
        return tile_polygon

    def _should_skip_tile(self, area: shpg.Polygon, tile: shpg.Polygon) -> bool:
        """
        Check if tile is outside area boundaries.

        If so, skip it.
        """
        intersects: bool = tile.intersects(area)
        return not intersects

    def _coordinates_to_x_y(self, latitude: float, longitude: float) -> tuple[int, int]:
        """
        Convert latitude and longitude into x and y values using `self.zoom`.

        Based on https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Implementations.
        """
        n_rows = 2**self.zoom
        x_tile = int(n_rows * ((longitude + 180) / 360))
        lat_radian = np.radians(latitude)
        y_tile = int((1 - np.arcsinh(np.tan(lat_radian)) / np.pi) / 2 * n_rows)
        return x_tile, y_tile

    def _x_y_to_coordinates(self, x: int, y: int) -> tuple[float, float]:
        """
        Convert x and y values into latitude and longitude using `self.zoom`.

        Based on https://wiki.openstreetmap.org/wiki/Slippy_map_tilenames#Implementations.
        """
        n = 2.0**self.zoom
        lon_deg = x / n * 360.0 - 180.0
        lat_rad = np.arctan(np.sinh(np.pi * (1 - 2 * y / n)))
        lat_deg = np.degrees(lat_rad)
        return (lat_deg, lon_deg)



"""
Hex2Vec model.

This module contains the embedding model from Hex2Vec paper[1].

References:
    [1] https://dl.acm.org/doi/10.1145/3486635.3491076
"""
from typing import TYPE_CHECKING, List, Tuple

from srai._optional import import_optional_dependencies
from srai.embedders import Model

if TYPE_CHECKING:  # pragma: no cover
    import torch


class Hex2VecModel(Model):
    """
    Hex2Vec embedding model.

    This class implements the embedding model from Hex2Vec paper. It is based on a skip-gram model
    with negative sampling and triplet-loss. The model takes vectors of numbers as input (raw counts
    of features) per region and outputs dense embeddings.
    """

    def __init__(self, layer_sizes: List[int], learning_rate: float = 0.001):
        """
        Initialize Hex2VecModel.

        Args:
            layer_sizes (List[int]): List of sizes for the model layers.
                The first element is the input size (number of features),
                the last element is the output (embedding) size.
            learning_rate (float, optional): Learning rate. Defaults to 0.001.


        Raises:
            ValueError: If layer_sizes contains less than 2 elements.
        """
        import_optional_dependencies(
            dependency_group="torch", modules=["torch", "pytorch_lightning"]
        )
        from torch import nn

        super().__init__()
        self.layer_sizes = layer_sizes
        self.learning_rate = learning_rate

        if len(layer_sizes) < 2:
            raise ValueError("layer_sizes must contain at least 2 elements")

        def create_layers(sizes: List[Tuple[int, int]]) -> nn.Sequential:
            layers = []
            for i, (input_size, output_size) in enumerate(sizes):
                linear = nn.Linear(input_size, output_size)
                nn.init.xavier_uniform_(linear.weight)
                layers.append(nn.Linear(input_size, output_size))
                if i != len(sizes) - 1:
                    layers.append(nn.ReLU())
            return nn.Sequential(*layers)

        sizes = list(zip(layer_sizes[:-1], layer_sizes[1:]))
        self.encoder = create_layers(sizes)

    def forward(self, X_anchor: "torch.Tensor") -> "torch.Tensor":
        """
        Calculate embedding for a region.

        Args:
            X_anchor (torch.Tensor): Region features.
        """
        return self.encoder(X_anchor)

    def predict_proba(self, X_anchor: "torch.Tensor", X_context: "torch.Tensor") -> "torch.Tensor":
        """
        Predict the probability of X_anchor being neighbours with X_context.

        X_anchor and X_context are assumed to have the same batch size.
        The probabilities are calculated in pairs, i.e. the first element of X_anchor
        is compared with the first element of X_context.

        Args:
            X_anchor (torch.Tensor): Anchor regions.
            X_context (torch.Tensor): Context regions.
        """
        from torch.nn.functional import sigmoid

        score = self.predict_scores(X_anchor, X_context)
        return sigmoid(score)

    def predict_scores(self, X_anchor: "torch.Tensor", X_context: "torch.Tensor") -> "torch.Tensor":
        """
        Predict raw unnormalized scores of X_anchor being neighbours with X_context.

        X_anchor and X_context are assumed to have the same batch size.
        The scores are calculated in pairs, i.e. the first element of X_anchor
        is compared with the first element of X_context.
        In order to get probabilities, use the sigmoid function.

        Args:
            X_anchor (torch.Tensor): Anchor regions.
            X_context (torch.Tensor): Context regions.
        """
        import torch

        X_anchor_em = self(X_anchor)
        X_context_em = self(X_context)
        score = torch.mul(X_anchor_em, X_context_em).sum(dim=1)
        return score

    def training_step(self, batch: List["torch.Tensor"], batch_idx: int) -> "torch.Tensor":
        """
        Perform one training step.

        One batch of data consists of 3 tensors:
                - X_anchor: Anchor regions.
                - X_positive: Positive regions. The regions assumed to be neighbours
                    of the corresponding regions in X_anchor.
                - X_negative: Negative regions. The regions assumed to NOT be neighbours
                    of the corresponding regions in X_anchor.
            The regions in X_anchor, X_positive and X_negative are first embedded using the encoder.
            After that, the dot product of the corresponding embeddings is calculated.
            The loss is calculated as a binary cross-entropy between the dot product and the labels.

        Args:
            batch (List[torch.Tensor]): Batch of data.
            batch_idx (int): Batch index.
        """
        import torch
        import torch.nn.functional as F
        from torchmetrics.functional import f1_score as f1

        X_anchor, X_positive, X_negative = batch
        scores_pos = self.predict_scores(X_anchor, X_positive)
        scores_neg = self.predict_scores(X_anchor, X_negative)

        scores = torch.cat([scores_pos, scores_neg])
        y_pos = torch.ones_like(scores_pos)
        y_neg = torch.zeros_like(scores_neg)
        y = torch.cat([y_pos, y_neg]).to(X_anchor)

        loss = F.binary_cross_entropy_with_logits(scores, y)
        f_score = f1(F.sigmoid(scores), y.int(), task="binary")
        self.log("train_loss", loss, on_step=True, on_epoch=True)
        self.log("train_f1", f_score, on_step=True, on_epoch=True)
        return loss

    def validation_step(self, batch: List["torch.Tensor"], batch_idx: int) -> "torch.Tensor":
        """
        Perform one validation step.

        Args:
            batch (List[torch.Tensor]): Batch of data.
            batch_idx (int): Batch index.
        """
        import torch
        import torch.nn.functional as F
        from torchmetrics.functional import f1_score as f1

        X_anchor, X_positive, X_negative = batch
        scores_pos = self.predict_scores(X_anchor, X_positive)
        scores_neg = self.predict_scores(X_anchor, X_negative)

        scores = torch.cat([scores_pos, scores_neg])
        y_pos = torch.ones_like(scores_pos)
        y_neg = torch.zeros_like(scores_neg)
        y = torch.cat([y_pos, y_neg]).to(X_anchor)

        loss = F.binary_cross_entropy_with_logits(scores, y)
        f_score = f1(F.sigmoid(scores), y.int(), task="binary")
        self.log("val_loss", loss, on_step=True, on_epoch=True)
        self.log("val_f1", f_score, on_step=True, on_epoch=True)
        return loss

    def configure_optimizers(self) -> "torch.optim.Optimizer":
        """Configure optimizer."""
        import torch

        return torch.optim.Adam(self.parameters(), lr=self.learning_rate)



"""
PBF File Clipper.

This module contains a clipper capable of clipping a PBF file to a smaller size using osmconvert or
osmosis CLI tools.
"""
import os
import shutil
import stat
import tempfile
import zipfile
from pathlib import Path
from sys import platform
from typing import Sequence, Union

import numpy as np
from shapely.geometry import MultiPolygon, Polygon
from tqdm import tqdm

from srai.geometry import get_geometry_hash
from srai.loaders.download import download_file


class PbfFileClipper:
    """
    PbfFileClipper.

    PBF(Protocolbuffer Binary Format)[1] file clipper is a tool
    capable of clipping `*.osm.pbf` files with OSM data for a given area to make it smaller.

    Class will automatically download those CLI tools and execute them from Python code.

    This clipper uses two tools to clip a PBF file for a given region:
     - Osmconvert [2] - used on Linux systems
     - Osmosis [3] - used on OS X and Windows systems

    References:
        1. https://wiki.openstreetmap.org/wiki/PBF_Format
        2. https://wiki.openstreetmap.org/wiki/Osmconvert
        3. https://wiki.openstreetmap.org/wiki/Osmosis
    """

    def __init__(self, working_directory: Union[str, Path] = "files") -> None:
        """
        Initialize PbfFileClipper.

        Args:
            working_directory (Union[str, Path], optional): Directory where to save
                the parsed `*.osm.pbf` files. Defaults to "files".
        """
        self.working_directory = Path(working_directory)

        # Tools for clipping OSM files into smaller regions
        # osmconvert is faster, but works on Linux
        self.OSMCONVERT_PATH = (self.working_directory / "osmconvert_tool/osmconvert").as_posix()
        # osmosis is slower, but also works on Mac OS
        self.OSMOSIS_DIRECTORY_PATH = (self.working_directory / "osmosis_tool").as_posix()
        self.OSMOSIS_EXECUTABLE_PATH = (
            self.working_directory / "osmosis_tool/bin/osmosis"
        ).as_posix()

    def clip_pbf_file(
        self,
        geometry: Polygon,
        pbf_files: Sequence[Union[str, "os.PathLike[str]"]],
    ) -> Path:
        """
        Clip a list of PBF files using a given geometry and merge them into a single file.

        Args:
            geometry (Polygon): Geometry to be used for clipping.
            pbf_files (Sequence[Union[str, os.PathLike[str]]]): List of PBF files to be processed.

        Raises:
            RuntimeError: If system is unrecognized (not one of linux, linux2, darwin or win32)

        Returns:
            Path: Location of merged PBF file.
        """
        geometry_hash = get_geometry_hash(geometry)
        final_osm_path = (self.working_directory.resolve() / f"{geometry_hash}.osm.pbf").as_posix()

        if Path(final_osm_path).exists():
            return Path(final_osm_path)

        with tempfile.TemporaryDirectory(dir=self.working_directory) as tmp_dir_name:
            tmp_dir_path = Path(tmp_dir_name)

            final_osm_path_alphanumeric_safe = (
                tmp_dir_path.resolve() / f"merged_{geometry_hash}.osm.pbf"
            ).as_posix()

            poly_path = (tmp_dir_path.resolve() / f"poly_files/{geometry_hash}.poly").as_posix()

            # create poly file
            self._generate_poly_file(geometry, poly_path)

            if platform in ("linux", "linux2"):
                download_function = self._download_osmconvert_tool
                tmp_file_extension = "o5m"
                clip_function = self._clip_pbf_with_osmconvert
                merge_function = self._merge_pbfs_with_osmconvert
            elif platform in ("darwin", "win32"):
                download_function = self._download_osmosis_tool
                tmp_file_extension = "osm.pbf"
                clip_function = self._clip_pbf_with_osmosis
                merge_function = self._merge_pbfs_with_osmosis
            else:
                raise RuntimeError(f"Unrecognized system platform: {platform}")

            download_function()

            # clip all pbf_files in temp directory
            clipped_pbf_files_set = set()
            for pbf_file_path in tqdm(pbf_files, desc="Clipping PBF files"):
                pbf_file_path_posix = Path(pbf_file_path).as_posix()
                pbf_file_name = Path(pbf_file_path).stem.split(".")[0]

                osm_path = (
                    tmp_dir_path.resolve()
                    / f"pbf_files/{pbf_file_name}_{geometry_hash}.{tmp_file_extension}"
                ).as_posix()

                clipped_pbf_files_set.add(osm_path)

                if Path(osm_path).exists():
                    continue

                clip_function(pbf_file_path_posix, poly_path, osm_path)

            clipped_pbf_files_list = list(clipped_pbf_files_set)

            # if single file - copy it
            if len(clipped_pbf_files_list) == 1:
                if tmp_file_extension == "o5m":
                    self._convert_o5m_to_pbf(clipped_pbf_files_list[0], final_osm_path)
                else:
                    shutil.copy(clipped_pbf_files_list[0], final_osm_path)
            # merge all of the files and copy it
            else:
                merge_function(clipped_pbf_files_list, final_osm_path_alphanumeric_safe)
                shutil.copy(final_osm_path_alphanumeric_safe, final_osm_path)

        return Path(final_osm_path)

    def _generate_poly_file(self, geometry: Union[Polygon, MultiPolygon], poly_file: str) -> None:
        """
        This function will create the .poly files from the nuts shapefile.

        These.poly files are used to extract data from the openstreetmap files. This function is
        adapted from the OSMPoly function in QGIS.
        """
        Path(poly_file).parent.mkdir(parents=True, exist_ok=True)

        # this will create a list of the different subpolygons
        if geometry.geom_type == "MultiPolygon":
            polygons = geometry.geoms

        # the list will be lenght 1 if it is just one polygon
        elif geometry.geom_type == "Polygon":
            polygons = [geometry]

        # start writing the .poly file
        f = open(poly_file, "w")
        f.write("polygon\n")

        i = 0

        # loop over the different polygons, get their exterior and write the
        # coordinates of the ring to the .poly file
        for polygon in polygons:
            polygon_coords = np.array(polygon.exterior.coords)

            j = 0
            f.write(str(i) + "\n")

            for ring in polygon_coords:
                j = j + 1
                f.write("    " + str(ring[0]) + "     " + str(ring[1]) + "\n")

            i = i + 1
            # close the ring of one subpolygon if done
            f.write("END\n")

        # close the file when done
        f.write("END\n")
        f.close()

    def _download_osmconvert_tool(self) -> None:
        file_path = Path(self.OSMCONVERT_PATH)
        if not file_path.exists():
            tool_url = "http://m.m.i24.cc/osmconvert64"
            download_file(tool_url, self.OSMCONVERT_PATH)

            # make it executable
            file_path.chmod(file_path.stat().st_mode | stat.S_IEXEC)

    def _download_osmosis_tool(self) -> None:
        zip_path = (Path(self.OSMOSIS_DIRECTORY_PATH) / "osmosis.zip").as_posix()
        file_path = Path(self.OSMOSIS_EXECUTABLE_PATH)
        if not file_path.exists():
            tool_url = "https://github.com/openstreetmap/osmosis/releases/download/0.48.3/osmosis-0.48.3.zip"
            download_file(tool_url, zip_path)

            with zipfile.ZipFile(zip_path, "r") as zf:
                for member in tqdm(zf.infolist(), desc=""):
                    try:
                        zf.extract(member, self.OSMOSIS_DIRECTORY_PATH)
                    except zipfile.error:
                        pass

            # make it executable
            file_path.chmod(file_path.stat().st_mode | stat.S_IEXEC)

    def _clip_pbf_with_osmconvert(
        self, source_pbf_path: str, poly_file_path: str, output_pbf_path: str
    ) -> None:
        """Wrapper over osmconvert CLI tool."""
        Path(output_pbf_path).parent.mkdir(parents=True, exist_ok=True)
        os.system(
            "{}  {} -B={} --complete-ways --complete-multipolygons -t={} -o={}".format(
                self.OSMCONVERT_PATH,
                source_pbf_path,
                poly_file_path,
                Path(output_pbf_path).parent.as_posix(),
                output_pbf_path,
            )
        )

    def _clip_pbf_with_osmosis(
        self, source_pbf_path: str, poly_file_path: str, output_pbf_path: str
    ) -> None:
        """Wrapper over osmosis CLI tool."""
        Path(output_pbf_path).parent.mkdir(parents=True, exist_ok=True)
        os.system(
            '{} --read-pbf file="{}" --bounding-polygon file="{}" completeWays=yes'
            ' completeRelations=yes --write-pbf file="{}"'.format(
                self.OSMOSIS_EXECUTABLE_PATH, source_pbf_path, poly_file_path, output_pbf_path
            )
        )

    # osmconvert file1.o5m file2.o5m -o=merged.pbf
    def _merge_pbfs_with_osmconvert(self, pbf_paths: Sequence[str], output_pbf_path: str) -> None:
        Path(output_pbf_path).parent.mkdir(parents=True, exist_ok=True)
        joined_files = " ".join(pbf_paths)
        command = f"{self.OSMCONVERT_PATH} {joined_files} -o={output_pbf_path}"
        os.system(command)

    # osmosis --rb file1.pbf --rb file2.pbf --rb file3.pbf --merge --merge --wb merged.pbf
    def _merge_pbfs_with_osmosis(self, pbf_paths: Sequence[str], output_pbf_path: str) -> None:
        Path(output_pbf_path).parent.mkdir(parents=True, exist_ok=True)
        joined_files = " ".join([f"--rb {pbf_path}" for pbf_path in pbf_paths])
        merge_commands = " ".join(["--merge" for _ in range(len(pbf_paths) - 1)])
        command = (
            f"{self.OSMOSIS_EXECUTABLE_PATH} {joined_files} {merge_commands} --wb {output_pbf_path}"
        )
        os.system(command)

    def _convert_o5m_to_pbf(self, o5m_path: Sequence[str], output_pbf_path: str) -> None:
        Path(output_pbf_path).parent.mkdir(parents=True, exist_ok=True)
        command = f"{self.OSMCONVERT_PATH} {o5m_path} -o={output_pbf_path}"
        os.system(command)



"""
Hex2Vec filter.

This module contains the OSM tags filter that was used in Hex2Vec [1].

References:
    1. https://dl.acm.org/doi/10.1145/3486635.3491076
"""
from srai.loaders.osm_loaders.filters import OsmTagsFilter

HEX2VEC_FILTER: OsmTagsFilter = {
    "aeroway": [
        "aerodrome",
        "apron",
        "gate",
        "hangar",
        "helipad",
        "heliport",
        "navigationaid",
        "runway",
        "taxiway",
        "terminal",
        "windsock",
    ],
    "amenity": [
        "animal_boarding",
        "animal_breeding",
        "animal_shelter",
        "arts_centre",
        "atm",
        "baby_hatch",
        "bank",
        "bar",
        "bbq",
        "bench",
        "bicycle_parking",
        "bicycle_rental",
        "bicycle_repair_station",
        "biergarten",
        "boat_rental",
        "boat_sharing",
        "brothel",
        "bureau_de_change",
        "bus_station",
        "cafe",
        "car_rental",
        "car_sharing",
        "car_wash",
        "casino",
        "charging_station",
        "childcare",
        "cinema",
        "clinic",
        "clock",
        "college",
        "community_centre",
        "conference_centre",
        "courthouse",
        "crematorium",
        "dentist",
        "dive_centre",
        "doctors",
        "dog_toilet",
        "drinking_water",
        "driving_school",
        "embassy",
        "events_venue",
        "fast_food",
        "ferry_terminal",
        "fire_station",
        "food_court",
        "fountain",
        "fuel",
        "funeral_hall",
        "gambling",
        "give_box",
        "grave_yard",
        "grit_bin",
        "gym",
        "hospital",
        "hunting_stand",
        "ice_cream",
        "internet_cafe",
        "kindergarten",
        "kitchen",
        "kneipp_water_cure",
        "language_school",
        "library",
        "lounger",
        "love_hotel",
        "marketplace",
        "monastery",
        "motorcycle_parking",
        "music_school",
        "nightclub",
        "nursing_home",
        "parking",
        "parking_entrance",
        "parking_space",
        "pharmacy",
        "photo_booth",
        "place_of_mourning",
        "place_of_worship",
        "planetarium",
        "police",
        "post_box",
        "post_depot",
        "post_office",
        "prison",
        "pub",
        "public_bath",
        "public_bookcase",
        "ranger_station",
        "recycling",
        "refugee_site",
        "restaurant",
        "sanitary_dump_station",
        "school",
        "shelter",
        "shower",
        "social_centre",
        "social_facility",
        "stripclub",
        "studio",
        "swingerclub",
        "taxi",
        "telephone",
        "theatre",
        "toilets",
        "townhall",
        "toy_library",
        "university",
        "vehicle_inspection",
        "vending_machine",
        "veterinary",
        "waste_disposal",
        "waste_transfer_station",
        "water_point",
        "watering_place",
    ],
    "building": [
        "apartments",
        "bakehouse",
        "barn",
        "bridge",
        "bungalow",
        "bunker",
        "cabin",
        "carport",
        "cathedral",
        "chapel",
        "church",
        "civic",
        "college",
        "commercial",
        "conservatory",
        "construction",
        "container",
        "cowshed",
        "detached",
        "dormitory",
        "farm",
        "farm_auxiliary",
        "fire_station",
        "garage",
        "garages",
        "gatehouse",
        "government",
        "grandstand",
        "greenhouse",
        "hangar",
        "hospital",
        "hotel",
        "house",
        "houseboat",
        "hut",
        "industrial",
        "kindergarten",
        "kiosk",
        "military",
        "monastery",
        "mosque",
        "office",
        "parking",
        "pavilion",
        "presbytery",
        "public",
        "religious",
        "residential",
        "retail",
        "riding_hall",
        "roof",
        "ruins",
        "school",
        "semidetached_house",
        "service",
        "shed",
        "shrine",
        "sports_hall",
        "stable",
        "stadium",
        "static_caravan",
        "sty",
        "supermarket",
        "synagogue",
        "temple",
        "terrace",
        "toilets",
        "train_station",
        "transformer_tower",
        "transportation",
        "tree_house",
        "university",
        "warehouse",
        "water_tower",
        "yes",
    ],
    "healthcare": [
        "alternative",
        "audiologist",
        "birthing_center",
        "blood_bank",
        "blood_donation",
        "counselling",
        "dialysis",
        "hospice",
        "laboratory",
        "midwife",
        "nurse",
        "occupational_therapist",
        "optometrist",
        "physiotherapist",
        "podiatrist",
        "psychotherapist",
        "rehabilitation",
        "sample_collection",
        "speech_therapist",
        "vaccination_centre",
    ],
    "historic": [
        "aircraft",
        "aqueduct",
        "archaeological_site",
        "battlefield",
        "bomb_crater",
        "boundary_stone",
        "building",
        "cannon",
        "castle",
        "castle_wall",
        "charcoal_pile",
        "church",
        "city_gate",
        "citywalls",
        "farm",
        "fort",
        "highwater_mark",
        "locomotive",
        "manor",
        "memorial",
        "milestone",
        "monastery",
        "monument",
        "pillory",
        "railway_car",
        "ruins",
        "rune_stone",
        "ship",
        "tank",
        "tower",
        "wayside_cross",
        "wayside_shrine",
        "wreck",
        "yes",
    ],
    "landuse": [
        "allotments",
        "basin",
        "brownfield",
        "cemetery",
        "commercial",
        "conservation",
        "construction",
        "depot",
        "farmland",
        "farmyard",
        "flowerbed",
        "forest",
        "garages",
        "greenfield",
        "greenhouse_horticulture",
        "industrial",
        "landfill",
        "meadow",
        "military",
        "orchard",
        "plant_nursery",
        "port",
        "quarry",
        "railway",
        "recreation_ground",
        "religious",
        "reservoir",
        "residential",
        "retail",
        "village_green",
        "vineyard",
        "winter_sports",
    ],
    "leisure": [
        "adult_gaming_centre",
        "amusement_arcade",
        "bandstand",
        "beach_resort",
        "bird_hide",
        "common",
        "dance",
        "disc_golf_course",
        "dog_park",
        "escape_game",
        "firepit",
        "fishing",
        "fitness_centre",
        "fitness_station",
        "garden",
        "hackerspace",
        "horse_riding",
        "ice_rink",
        "marina",
        "miniature_golf",
        "nature_reserve",
        "park",
        "picnic_table",
        "pitch",
        "playground",
        "slipway",
        "sports_centre",
        "stadium",
        "summer_camp",
        "swimming_area",
        "swimming_pool",
        "track",
        "water_park",
    ],
    "military": [
        "airfield",
        "barracks",
        "bunker",
        "checkpoint",
        "danger_area",
        "naval_base",
        "obstacle_course",
        "office",
        "range",
        "training_area",
        "trench",
    ],
    "natural": [
        "arete",
        "bare_rock",
        "bay",
        "beach",
        "cape",
        "cave_entrance",
        "cliff",
        "coastline",
        "dune",
        "fell",
        "grassland",
        "heath",
        "hot_spring",
        "isthmus",
        "mud",
        "peak",
        "peninsula",
        "reef",
        "ridge",
        "rock",
        "saddle",
        "sand",
        "scree",
        "scrub",
        "shingle",
        "sinkhole",
        "spring",
        "stone",
        "strait",
        "water",
        "wetland",
        "wood",
    ],
    "office": [
        "accountant",
        "advertising_agency",
        "architect",
        "association",
        "charity",
        "company",
        "consulting",
        "courier",
        "coworking",
        "diplomatic",
        "educational_institution",
        "employment_agency",
        "energy_supplier",
        "engineer",
        "estate_agent",
        "financial",
        "financial_advisor",
        "forestry",
        "foundation",
        "government",
        "guide",
        "insurance",
        "it",
        "lawyer",
        "logistics",
        "moving_company",
        "newspaper",
        "ngo",
        "notary",
        "political_party",
        "property_management",
        "quango",
        "religion",
        "research",
        "surveyor",
        "tax_advisor",
        "telecommunication",
        "visa",
        "water_utility",
        "yes",
    ],
    "shop": [
        "agrarian",
        "alcohol",
        "anime",
        "antiques",
        "appliance",
        "art",
        "atv",
        "baby_goods",
        "bag",
        "bakery",
        "bathroom_furnishing",
        "beauty",
        "bed",
        "beverages",
        "bicycle",
        "boat",
        "bookmaker",
        "books",
        "boutique",
        "brewing_supplies",
        "butcher",
        "camera",
        "candles",
        "cannabis",
        "car",
        "car_parts",
        "car_repair",
        "caravan",
        "carpet",
        "charity",
        "cheese",
        "chemist",
        "chocolate",
        "clothes",
        "coffee",
        "collector",
        "computer",
        "confectionery",
        "convenience",
        "copyshop",
        "cosmetics",
        "craft",
        "curtain",
        "dairy",
        "deli",
        "department_store",
        "doityourself",
        "doors",
        "drugstore",
        "dry_cleaning",
        "electrical",
        "electronics",
        "energy",
        "erotic",
        "fabric",
        "farm",
        "fashion",
        "fashion_accessories",
        "fireplace",
        "fishing",
        "flooring",
        "florist",
        "frame",
        "frozen_food",
        "fuel",
        "funeral_directors",
        "furniture",
        "games",
        "garden_centre",
        "garden_furniture",
        "gas",
        "general",
        "gift",
        "glaziery",
        "golf",
        "greengrocer",
        "groundskeeping",
        "hairdresser",
        "hairdresser_supply",
        "hardware",
        "health_food",
        "hearing_aids",
        "herbalist",
        "hifi",
        "household_linen",
        "houseware",
        "hunting",
        "ice_cream",
        "interior_decoration",
        "jewelry",
        "kiosk",
        "kitchen",
        "lamps",
        "laundry",
        "leather",
        "lighting",
        "locksmith",
        "lottery",
        "mall",
        "massage",
        "medical_supply",
        "military_surplus",
        "mobile_phone",
        "model",
        "money_lender",
        "motorcycle",
        "music",
        "musical_instrument",
        "newsagent",
        "nutrition_supplements",
        "optician",
        "organic",
        "outdoor",
        "outpost",
        "paint",
        "party",
        "pasta",
        "pastry",
        "pawnbroker",
        "perfumery",
        "pest_control",
        "pet",
        "pet_grooming",
        "photo",
        "pyrotechnics",
        "radiotechnics",
        "religion",
        "scuba_diving",
        "seafood",
        "second_hand",
        "security",
        "sewing",
        "shoes",
        "ski",
        "spices",
        "sports",
        "stationery",
        "storage_rental",
        "supermarket",
        "swimming_pool",
        "tailor",
        "tattoo",
        "tea",
        "ticket",
        "tiles",
        "tobacco",
        "toys",
        "trade",
        "travel_agency",
        "trophy",
        "tyres",
        "vacant",
        "vacuum_cleaner",
        "variety_store",
        "video",
        "video_games",
        "watches",
        "water",
        "weapons",
        "wholesale",
        "window_blind",
        "windows",
        "wine",
        "wool",
    ],
    "sport": [
        "10pin",
        "9pin",
        "aikido",
        "american_football",
        "archery",
        "athletics",
        "badminton",
        "bandy",
        "baseball",
        "basketball",
        "beachvolleyball",
        "biathlon",
        "billiards",
        "bmx",
        "bobsleigh",
        "boules",
        "bowls",
        "boxing",
        "bullfighting",
        "canadian_football",
        "canoe",
        "chess",
        "cliff_diving",
        "climbing",
        "climbing_adventure",
        "cricket",
        "croquet",
        "crossfit",
        "curling",
        "cycling",
        "darts",
        "dog_agility",
        "dog_racing",
        "equestrian",
        "fencing",
        "field_hockey",
        "fitness",
        "floorball",
        "free_flying",
        "futsal",
        "gaelic_games",
        "golf",
        "gymnastics",
        "handball",
        "hapkido",
        "horse_racing",
        "horseshoes",
        "ice_hockey",
        "ice_skating",
        "ice_stock",
        "judo",
        "karate",
        "karting",
        "kickboxing",
        "kitesurfing",
        "korfball",
        "lacrosse",
        "martial_arts",
        "miniature_golf",
        "model_aerodrome",
        "motocross",
        "motor",
        "multi",
        "netball",
        "obstacle_course",
        "orienteering",
        "paddle_tennis",
        "padel",
        "parachuting",
        "parkour",
        "pelota",
        "pickleball",
        "pilates",
        "pole_dance",
        "racquet",
        "rc_car",
        "roller_skating",
        "rowing",
        "rugby_league",
        "rugby_union",
        "running",
        "sailing",
        "scuba_diving",
        "shooting",
        "skateboard",
        "ski_jumping",
        "skiing",
        "snooker",
        "soccer",
        "speedway",
        "squash",
        "surfing",
        "swimming",
        "table_soccer",
        "table_tennis",
        "taekwondo",
        "tennis",
        "toboggan",
        "volleyball",
        "wakeboarding",
        "water_polo",
        "water_ski",
        "weightlifting",
        "wrestling",
        "yoga",
    ],
    "tourism": [
        "alpine_hut",
        "apartment",
        "aquarium",
        "artwork",
        "attraction",
        "camp_pitch",
        "camp_site",
        "caravan_site",
        "chalet",
        "gallery",
        "guest_house",
        "hostel",
        "hotel",
        "information",
        "motel",
        "museum",
        "picnic_site",
        "theme_park",
        "viewpoint",
        "wilderness_hut",
        "yes",
        "zoo",
    ],
    "water": [
        "basin",
        "canal",
        "ditch",
        "fish_pass",
        "lagoon",
        "lake",
        "lock",
        "moat",
        "oxbow",
        "pond",
        "reflecting_pool",
        "reservoir",
        "river",
        "wastewater",
    ],
    "waterway": [
        "boatyard",
        "canal",
        "dam",
        "ditch",
        "dock",
        "drain",
        "fairway",
        "fuel",
        "lock_gate",
        "pressurised",
        "river",
        "riverbank",
        "stream",
        "tidal_channel",
        "turning_point",
        "water_point",
        "waterfall",
        "weir",
    ],
}



"""Tests for DataCollector subclasses."""
import os
from pathlib import Path
from typing import Union

import PIL
import pytest
from numpy.random import default_rng
from pytest_mock import MockerFixture

import srai.loaders.osm_loaders.osm_tile_data_collector as collectors

rng = default_rng()
PATH = "path"
FILE_TYPE = "png"


def create_id(x: int, y: int) -> str:
    """Create test id."""
    return f"{x}_{y}_ZOOM"


class TestSavingDataCollector:
    """Tests for class SavingDataCollector."""

    @pytest.fixture  # type: ignore
    def col(self) -> collectors.SavingDataCollector:
        """Fixture for SavingDataCollector."""
        return collectors.SavingDataCollector(PATH, FILE_TYPE)

    def test_save_to_disk(self, mocker: MockerFixture, col: collectors.SavingDataCollector) -> None:
        """Test if save to disk saves image properly."""
        _path_image_save(mocker)
        x, y = 1, 1
        expected = _get_expected_path(x, y)

        path = col.store(
            create_id(x, y), PIL.Image.fromarray(rng.integers(0, 256, size=(3, 3), dtype="uint8"))
        )

        PIL.Image.Image.save.assert_called_once_with(expected)
        assert _get_expected_path(x, y) == path


def _get_expected_path(x: int, y: int) -> Path:
    return Path(os.path.join(PATH, f"{create_id(x,y)}.{FILE_TYPE}"))


def _path_image_save(mocker: MockerFixture) -> None:
    mocker.patch("PIL.Image.Image.save")


class TestInMemoryDataCollector:
    """Test for class InMemoryDataCollector."""

    @pytest.fixture  # type: ignore
    def col(self) -> collectors.InMemoryDataCollector:
        """Fixture for InMemoryDataCollector."""
        return collectors.InMemoryDataCollector()

    def test_should_return_stored(self, col: collectors.InMemoryDataCollector) -> None:
        """Test values of collected images."""
        x, y = 1, 1
        img = PIL.Image.fromarray(rng.integers(0, 256, size=(3, 3), dtype="uint8"))

        stored = col.store(create_id(x, y), img)

        assert stored == img


@pytest.mark.parametrize(  # type: ignore
    "collector_type", [(collectors.DataCollectorType.RETURN), "return"]
)
def test_in_memory_collector_creation(
    collector_type: Union[str, collectors.DataCollectorType]
) -> None:
    """Test if factory creates properly InMemoryDataCollector."""
    created = collectors.get_collector(collector_type)

    assert isinstance(created, collectors.InMemoryDataCollector)


@pytest.mark.parametrize(  # type: ignore
    "collector_type", [(collectors.DataCollectorType.SAVE), "save"]
)
def test_saving_collector_creation(
    collector_type: Union[str, collectors.DataCollectorType]
) -> None:
    """Test if factory creates properly SavingDataCollector."""
    created = collectors.get_collector(collector_type, save_path=PATH, file_extension=FILE_TYPE)

    assert isinstance(created, collectors.SavingDataCollector), f"Invalid type {type(created)}"
    assert created.format == FILE_TYPE
    assert str(created.save_path) == PATH


def test_invalid_type() -> None:
    """Tests if throws on unknown type."""
    with pytest.raises(ValueError):
        collectors.get_collector("Some weird type")
